//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Coord/AxisNames.cpp
//! @brief     Implements functions in namespace AxisNames.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Device/Coord/AxisNames.h"
#include <map>

namespace DataUtils::AxisNames {

// For spherical detectors in GISAS simulations
BA_DEVICE_API_ const std::map<Axes::Coords, std::string> sphericalAxis0 = {
    {Axes::Coords::NBINS, "X [nbins]"},
    {Axes::Coords::RADIANS, "phi_f [rad]"},
    {Axes::Coords::DEGREES, "phi_f [deg]"},
    {Axes::Coords::QSPACE, "Qy [1/nm]"},
    {Axes::Coords::QXQY, "Qx [1/nm]"}};
BA_DEVICE_API_ const std::map<Axes::Coords, std::string> sphericalAxis1 = {
    {Axes::Coords::NBINS, "Y [nbins]"},
    {Axes::Coords::RADIANS, "alpha_f [rad]"},
    {Axes::Coords::DEGREES, "alpha_f [deg]"},
    {Axes::Coords::QSPACE, "Qz [1/nm]"},
    {Axes::Coords::QXQY, "Qy [1/nm]"}};

// For rectangular detectors in GISAS simulations
BA_DEVICE_API_ const std::map<Axes::Coords, std::string> rectangularAxis0 = {
    {Axes::Coords::NBINS, "X [nbins]"},     {Axes::Coords::RADIANS, "phi_f [rad]"},
    {Axes::Coords::DEGREES, "phi_f [deg]"}, {Axes::Coords::MM, "X [mm]"},
    {Axes::Coords::QSPACE, "Qy [1/nm]"},    {Axes::Coords::QXQY, "Qx [1/nm]"}};
BA_DEVICE_API_ const std::map<Axes::Coords, std::string> rectangularAxis1 = {
    {Axes::Coords::NBINS, "Y [nbins]"},       {Axes::Coords::RADIANS, "alpha_f [rad]"},
    {Axes::Coords::DEGREES, "alpha_f [deg]"}, {Axes::Coords::MM, "Y [mm]"},
    {Axes::Coords::QSPACE, "Qz [1/nm]"},      {Axes::Coords::QXQY, "Qy [1/nm]"}};

// For off-specular simulations (both spherical and rectangular detectors)
// Currently 'mm' is not supported for the y-axis
BA_DEVICE_API_ const std::map<Axes::Coords, std::string> offSpecularAxis0 = {
    {Axes::Coords::NBINS, "X [nbins]"},
    {Axes::Coords::RADIANS, "alpha_i [rad]"},
    {Axes::Coords::DEGREES, "alpha_i [deg]"}};
BA_DEVICE_API_ const std::map<Axes::Coords, std::string> offSpecularAxis1 = {
    {Axes::Coords::NBINS, "Y [nbins]"},
    {Axes::Coords::RADIANS, "alpha_f [rad]"},
    {Axes::Coords::DEGREES, "alpha_f [deg]"}};

BA_DEVICE_API_ const std::map<Axes::Coords, std::string> specAxis = {
    {Axes::Coords::NBINS, "X [nbins]"},
    {Axes::Coords::RADIANS, "alpha_i [rad]"},
    {Axes::Coords::DEGREES, "alpha_i [deg]"},
    {Axes::Coords::QSPACE, "Q [1/nm]"},
    {Axes::Coords::RQ4, "Q [1/nm]"}};

BA_DEVICE_API_ const std::map<Axes::Coords, std::string> specAxisQ = {
    {Axes::Coords::NBINS, "X [nbins]"},
    {Axes::Coords::QSPACE, "Q [1/nm]"},
    {Axes::Coords::RQ4, "Q [1/nm]"}};

// corner case: axis corresponding to the position
// across sample is accepts only bins and nanometers,
// but the units specified in dedicated unit
// converter correspond to inclination angular axis.
// For this reason depth axis map returns always
// nanometers except for bins.

BA_DEVICE_API_ const std::map<Axes::Coords, std::string> sampleDepthAxis = {
    {Axes::Coords::NBINS, "Y [nbins]"},
    {Axes::Coords::RADIANS, "Position [nm]"},
    {Axes::Coords::DEGREES, "Position [nm]"},
    {Axes::Coords::QSPACE, "Position [nm]"}};

} // namespace DataUtils::AxisNames
