//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Fit/FitParameterContainerItem.cpp
//! @brief     Implements class FitParameterContainerItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Fit/FitParameterContainerItem.h"
#include "Base/Util/Assert.h"
#include "Fit/Param/Parameters.h"
#include "GUI/Model/Fit/FitParameterItem.h"
#include "GUI/Model/Fit/FitParameterLinkItem.h"
#include "GUI/Model/Fit/ParameterTreeItems.h"
#include "GUI/Model/Session/ModelPath.h"
#include "GUI/Util/Error.h"

FitParameterContainerItem::FitParameterContainerItem() : SessionItem(M_TYPE)
{
    registerTag(T_FIT_PARAMETERS, 0, -1, {FitParameterItem::M_TYPE});
    setDefaultTag(T_FIT_PARAMETERS);
}

//! returns FitParameterItem for given link (path in model)

FitParameterItem* FitParameterContainerItem::fitParameterItem(const QString& link)
{
    for (FitParameterItem* item : items<FitParameterItem>(T_FIT_PARAMETERS)) {
        for (FitParameterLinkItem* linkItem : item->linkItems()) {
            if (link == linkItem->link())
                return dynamic_cast<FitParameterItem*>(item);
        }
    }
    return nullptr;
}

QVector<FitParameterItem*> FitParameterContainerItem::fitParameterItems()
{
    return items<FitParameterItem>(T_FIT_PARAMETERS);
}

bool FitParameterContainerItem::isEmpty()
{
    return getItems(T_FIT_PARAMETERS).isEmpty();
}

//! Propagate values to the corresponding parameter tree items of parameterContainer.

void FitParameterContainerItem::setValuesInParameterContainer(
    const QVector<double>& values, ParameterContainerItem* parameterContainer)
{
    ASSERT(parameterContainer);

    QVector<FitParameterItem*> fitPars = items<FitParameterItem>(T_FIT_PARAMETERS);

    int index(0);
    for (int i = 0; i < fitPars.size(); ++i) {
        auto link_list = fitPars[i]->linkItems();
        if (link_list.empty())
            continue;
        for (FitParameterLinkItem* linkItem : link_list) {
            QString parPath = linkItem->link();
            auto* itemInTuningTree = dynamic_cast<ParameterItem*>(
                GUI::Model::Path::getItemFromPath(parPath, parameterContainer));
            if (itemInTuningTree)
                itemInTuningTree->propagateValueToLink(values[index]);
        }
        index++;
    }
}

mumufit::Parameters FitParameterContainerItem::createParameters() const
{
    mumufit::Parameters result;

    int index(0);
    for (auto* item : getItems(FitParameterContainerItem::T_FIT_PARAMETERS)) {
        auto* fitPar = dynamic_cast<FitParameterItem*>(item);
        if (!fitPar->isValid()) {
            std::stringstream ss;
            ss << "FitParameterContainerItem::createParameters(): invalid starting value "
                  "or (min, max) range in fitting parameter par"
               << index;
            throw Error(QString::fromStdString(ss.str()));
        }
        double startValue = fitPar->startValue();
        AttLimits limits = fitPar->attLimits();
        QString name = QString("par%1").arg(index);
        result.add(mumufit::Parameter(name.toStdString(), startValue, limits));
        ++index;
    }

    return result;
}
