//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/From/GUIObjectBuilder.cpp
//! @brief     Implements GUI::Model::ObjectBuilder namespace
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/From/GUIObjectBuilder.h"
#include "Base/Const/Units.h"
#include "Base/Util/Assert.h"
#include "Core/Simulation/includeSimulations.h"
#include "GUI/Model/From/FromDomain.h"
#include "GUI/Model/From/GUIDomainSampleVisitor.h"
#include "GUI/Model/Instrument/InstrumentItems.h"
#include "GUI/Model/Instrument/InstrumentModel.h"
#include "GUI/Model/Session/SimulationOptionsItem.h"
#include "Resample/Options/SimulationOptions.h"

namespace {

GISASInstrumentItem* createGISASInstrumentItem(InstrumentModel* model,
                                               const GISASSimulation& simulation,
                                               const QString& name)
{
    auto* result = model->insertItem<GISASInstrumentItem>();
    result->setItemName(name);
    GUI::Transform::FromDomain::setGISASBeamItem(result->beamItem(), simulation);
    GUI::Transform::FromDomain::setDetector(result, simulation);
    GUI::Transform::FromDomain::setBackground(result, simulation);

    return result;
}

OffSpecularInstrumentItem* createOffSpecularInstrumentItem(InstrumentModel* model,
                                                           const OffSpecularSimulation& simulation,
                                                           const QString& name)
{
    auto* result = model->insertItem<OffSpecularInstrumentItem>();
    result->setItemName(name);
    GUI::Transform::FromDomain::setOffSpecularBeamItem(result->beamItem(), simulation);
    GUI::Transform::FromDomain::setDetector(result, simulation);
    GUI::Transform::FromDomain::setBackground(result, simulation);

    BasicAxisItem* axisItem = result->alphaAxis();
    ASSERT(axisItem);
    GUI::Transform::FromDomain::setAxisItem(axisItem, *simulation.beamAxis(), 1. / Units::deg);

    return result;
}

SpecularInstrumentItem* createSpecularInstrumentItem(InstrumentModel* model,
                                                     const SpecularSimulation& simulation,
                                                     const QString& name)
{
    auto* result = model->insertItem<SpecularInstrumentItem>();
    result->setItemName(name);
    GUI::Transform::FromDomain::setSpecularBeamItem(result->beamItem(), simulation);
    GUI::Transform::FromDomain::setBackground(result, simulation);

    return result;
}

} // namespace

//  ************************************************************************************************
//  class implementation
//  ************************************************************************************************

SessionItem* GUI::Model::ObjectBuilder::populateInstrumentModel(InstrumentModel* p_instrument_model,
                                                                const ISimulation& simulation,
                                                                const QString& instrument_name)
{
    ASSERT(p_instrument_model);

    QString name = instrument_name.isEmpty()
                       ? QString::fromStdString(simulation.instrument().className())
                       : instrument_name;

    if (const auto* gisasSimulation = dynamic_cast<const GISASSimulation*>(&simulation))
        return createGISASInstrumentItem(p_instrument_model, *gisasSimulation, name);
    if (const auto* offSpecSimulation = dynamic_cast<const OffSpecularSimulation*>(&simulation))
        return createOffSpecularInstrumentItem(p_instrument_model, *offSpecSimulation, name);
    if (const auto* spec_simulation = dynamic_cast<const SpecularSimulation*>(&simulation))
        return createSpecularInstrumentItem(p_instrument_model, *spec_simulation, name);
    ASSERT(0);
}

void GUI::Model::ObjectBuilder::populateSimulationOptions(SimulationOptionsItem* item,
                                                          const ISimulation& simulation)
{
    if (simulation.options().isIntegrate())
        item->setUseMonteCarloIntegration(
            static_cast<unsigned>(simulation.options().getMcPoints()));
    else
        item->setUseAnalytical();

    item->setUseAverageMaterials(simulation.options().useAvgMaterials());
    item->setIncludeSpecularPeak(simulation.options().includeSpecular());
}
