//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Group/ItemCatalog.cpp
//! @brief     Implements class ItemCatalog
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Group/ItemCatalog.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Data/Data1DViewItem.h"
#include "GUI/Model/Data/DataProperties.h"
#include "GUI/Model/Data/DataPropertyContainer.h"
#include "GUI/Model/Data/IntensityDataItem.h"
#include "GUI/Model/Data/MaskItems.h"
#include "GUI/Model/Data/PointwiseAxisItem.h"
#include "GUI/Model/Data/ProjectionItems.h"
#include "GUI/Model/Data/RealDataItem.h"
#include "GUI/Model/Data/SpecularDataItem.h"
#include "GUI/Model/Fit/FitParameterContainerItem.h"
#include "GUI/Model/Fit/FitParameterItem.h"
#include "GUI/Model/Fit/FitParameterLinkItem.h"
#include "GUI/Model/Fit/FitSuiteItem.h"
#include "GUI/Model/Fit/MinimizerItem.h"
#include "GUI/Model/Fit/ParameterTreeItems.h"
#include "GUI/Model/Group/PropertyItem.h"
#include "GUI/Model/Instrument/BackgroundItems.h"
#include "GUI/Model/Instrument/BeamAngleItems.h"
#include "GUI/Model/Instrument/BeamWavelengthItem.h"
#include "GUI/Model/Instrument/FootprintItems.h"
#include "GUI/Model/Instrument/InstrumentItems.h"
#include "GUI/Model/Instrument/RectangularDetectorItem.h"
#include "GUI/Model/Instrument/ResolutionFunctionItems.h"
#include "GUI/Model/Instrument/SpecularBeamInclinationItem.h"
#include "GUI/Model/Instrument/SphericalDetectorItem.h"
#include "GUI/Model/Job/JobItem.h"
#include "GUI/Model/Sample/FTDecayFunctionItems.h"
#include "GUI/Model/Sample/FTDistributionItems.h"
#include "GUI/Model/Sample/FormFactorItems.h"
#include "GUI/Model/Sample/InterferenceItems.h"
#include "GUI/Model/Sample/Lattice2DItems.h"
#include "GUI/Model/Sample/LayerItem.h"
#include "GUI/Model/Sample/LayerRoughnessItems.h"
#include "GUI/Model/Sample/MesoCrystalItem.h"
#include "GUI/Model/Sample/MultiLayerItem.h"
#include "GUI/Model/Sample/ParticleCompositionItem.h"
#include "GUI/Model/Sample/ParticleCoreShellItem.h"
#include "GUI/Model/Sample/ParticleItem.h"
#include "GUI/Model/Sample/ParticleLayoutItem.h"
#include "GUI/Model/Session/RealLimitsItems.h"
#include "GUI/Model/Session/SimulationOptionsItem.h"
#include "GUI/Model/Trafo/RotationItems.h"
#include "GUI/Model/Trafo/TransformationItem.h"
#include "GUI/Model/Types/VectorItem.h"
#include "GUI/Util/Error.h"

ItemCatalog::ItemCatalog()
{
    addItem<MultiLayerItem>();
    addItem<LayerItem>();
    addItem<ParticleLayoutItem>();
    addItem<ParticleItem>();
    addItem<TransformationItem>();

    addItem<ParticleCoreShellItem>("Core shell particle",
                                   ":/SampleDesignerToolbox/images/ParticleCoreShell_64x64.png",
                                   "A particle with a core/shell geometry");

    addItem<ParticleCompositionItem>("Particle Composition",
                                     ":/SampleDesignerToolbox/images/ParticleComposition_64x64.png",
                                     "Composition of particles with fixed positions");

    addItem<MesoCrystalItem>("Meso Crystal", ":/SampleDesignerToolbox/images/Mesocrystal_64x64.png",
                             "A 3D crystal structure of nanoparticles");


    addItem<Interference1DLatticeItem>("1D lattice", ":/SampleDesignerToolbox/images/Lattice1D.png",
                                       "Interference function of 1D lattice");

    addItem<Interference2DLatticeItem>("2D lattice", ":/SampleDesignerToolbox/images/Lattice2D.png",
                                       "Interference function of 2D lattice");

    addItem<Interference2DParaCrystalItem>("2D paracrystal",
                                           ":/SampleDesignerToolbox/images/ParaCrystal2D.png",
                                           "Interference function of two-dimensional paracrystal");

    addItem<InterferenceFinite2DLatticeItem>("Finite 2D lattice",
                                             ":/SampleDesignerToolbox/images/Lattice2DFinite.png",
                                             "Interference function of finite 2D lattice");

    addItem<InterferenceHardDiskItem>("Hard disk Percus-Yevick",
                                      ":/SampleDesignerToolbox/images/Lattice2D.png",
                                      "Interference function for hard disk Percus-Yevick");

    addItem<InterferenceRadialParaCrystalItem>("Radial paracrystal",
                                               ":/SampleDesignerToolbox/images/ParaCrystal1D.png",
                                               "Interference function of radial paracrystal");


    addItem<GISASInstrumentItem>();
    addItem<OffSpecularInstrumentItem>();
    addItem<SpecularInstrumentItem>();
    addItem<GISASBeamItem>();
    addItem<SpecularBeamItem>();
    addItem<NoBackgroundItem>();
    addItem<ConstantBackgroundItem>();
    addItem<PoissonNoiseBackgroundItem>();
    addItem<FootprintNoneItem>();
    addItem<FootprintGaussianItem>();
    addItem<FootprintSquareItem>();
    addItem<VectorItem>();
    addItem<PropertyItem>();

    addItem<AnisoPyramidItem>("Anisotropic pyramid",
                              ":/SampleDesignerToolbox/images/ff_AnisoPyramid_64x64.png",
                              "Pyramid with a rectangular base");

    addItem<BarGaussItem>("BarGauss", ":/SampleDesignerToolbox/images/ff_Box_64x64.png",
                          "Rectangular cuboid");

    addItem<BarLorentzItem>("BarLorentz", ":/SampleDesignerToolbox/images/ff_Box_64x64.png",
                            "Rectangular cuboid");

    addItem<BoxItem>("Box", ":/SampleDesignerToolbox/images/ff_Box_64x64.png",
                     "Rectangular cuboid");

    addItem<ConeItem>("Cone", ":/SampleDesignerToolbox/images/ff_Cone_64x64.png", "Truncated cone");

    addItem<Cone6Item>("Cone6", ":/SampleDesignerToolbox/images/ff_Cone6_64x64.png",
                       "A truncated pyramid, based on a regular hexagon");

    addItem<CuboctahedronItem>(
        "Cuboctahedron", ":/SampleDesignerToolbox/images/ff_Cuboctahedron_64x64.png",
        "Combination of two pyramids of different heights, sharing a common face");

    addItem<DodecahedronItem>(
        "Dodecahedron", ":/SampleDesignerToolbox/images/ff_Dodecahedron_64x64.png", "Dodecahedron");

    addItem<CylinderItem>("Cylinder", ":/SampleDesignerToolbox/images/ff_Cylinder_64x64.png",
                          "Cylinder with a circular base");

    addItem<EllipsoidalCylinderItem>(
        "Ellipsoidal cylinder", ":/SampleDesignerToolbox/images/ff_EllipsoidalCylinder_64x64.png",
        "Cylinder with an ellipse cross section");

    addItem<FullSphereItem>("Full sphere", ":/SampleDesignerToolbox/images/ff_FullSphere_64x64.png",
                            "Full sphere");

    addItem<FullSpheroidItem>("Full spheroid",
                              ":/SampleDesignerToolbox/images/ff_FullSpheroid_64x64.png",
                              "Ellipsoid of revolution");

    addItem<HemiEllipsoidItem>("Hemi ellipsoid",
                               ":/SampleDesignerToolbox/images/ff_HemiEllipsoid_64x64.png",
                               "A horizontally oriented ellipsoid, truncated at the central plane");

    addItem<IcosahedronItem>(
        "Icosahedron", ":/SampleDesignerToolbox/images/ff_Icosahedron_64x64.png", "Icosahedron");

    addItem<Prism3Item>("Prism3", ":/SampleDesignerToolbox/images/ff_Prism3_64x64.png",
                        "Prism with an equilateral triangle base");

    addItem<Prism6Item>("Prism6", ":/SampleDesignerToolbox/images/ff_Prism6_64x64.png",
                        "Prism with a regular hexagonal base");

    addItem<PyramidItem>("Pyramid", ":/SampleDesignerToolbox/images/ff_Pyramid_64x64.png",
                         "Truncated pyramid with a square base");

    addItem<CosineRippleBoxItem>("CosineRippleBox",
                                 ":/SampleDesignerToolbox/images/ff_Ripple1_64x64.png",
                                 "Particle with a cosine profile and a rectangular base");

    addItem<CosineRippleGaussItem>("CosineRippleGauss",
                                   ":/SampleDesignerToolbox/images/ff_Ripple1_64x64.png",
                                   "Particle with a cosine profile and a rectangular base");

    addItem<CosineRippleLorentzItem>("CosineRippleLorentz",
                                     ":/SampleDesignerToolbox/images/ff_Ripple1_64x64.png",
                                     "Particle with a cosine profile and a rectangular base");

    addItem<SawtoothRippleBoxItem>(
        "SawtoothRippleBox", ":/SampleDesignerToolbox/images/ff_Ripple2_64x64.png",
        "Particle with an asymmetric triangle profile and a rectangular base");

    addItem<SawtoothRippleGaussItem>(
        "SawtoothRippleGauss", ":/SampleDesignerToolbox/images/ff_Ripple2_64x64.png",
        "Particle with an asymmetric triangle profile and a rectangular base");

    addItem<SawtoothRippleLorentzItem>(
        "SawtoothRippleLorentz", ":/SampleDesignerToolbox/images/ff_Ripple2_64x64.png",
        "Particle with an asymmetric triangle profile and a rectangular base");

    addItem<TetrahedronItem>(
        "Tetrahedron", ":/SampleDesignerToolbox/images/ff_Tetrahedron_64x64.png",
        "Truncated polyhedron with equilateral triangle base and cropped side faces");

    addItem<TruncatedCubeItem>("Truncated cube",
                               ":/SampleDesignerToolbox/images/ff_TruncatedCube_64x64.png",
                               "A cube whose eight vertices have been removed");

    addItem<TruncatedSphereItem>("Truncated sphere",
                                 ":/SampleDesignerToolbox/images/ff_TruncatedSphere_64x64.png",
                                 "Spherical dome");

    addItem<TruncatedSpheroidItem>("Truncated spheroid",
                                   ":/SampleDesignerToolbox/images/ff_TruncatedSpheroid_64x64.png",
                                   "Spheroidal dome");

    addItem<XRotationItem>();
    addItem<YRotationItem>();
    addItem<ZRotationItem>();
    addItem<EulerRotationItem>();

    addItem<LayerBasicRoughnessItem>();
    addItem<LayerZeroRoughnessItem>();

    addItem<SphericalDetectorItem>();
    addItem<RectangularDetectorItem>();

    addItem<DistributionNoneItem>();
    addItem<DistributionGateItem>();
    addItem<DistributionLorentzItem>();
    addItem<DistributionGaussianItem>();
    addItem<DistributionLogNormalItem>();
    addItem<DistributionCosineItem>();
    addItem<DistributionTrapezoidItem>();

    addItem<FTDistribution1DCauchyItem>("Cauchy 1D");
    addItem<FTDistribution1DGaussItem>("Gauss 1D");
    addItem<FTDistribution1DGateItem>("Gate 1D");
    addItem<FTDistribution1DTriangleItem>("Triangle 1D");
    addItem<FTDistribution1DCosineItem>("Cosine 1D");
    addItem<FTDistribution1DVoigtItem>("Voigt 1D");

    addItem<FTDistribution2DCauchyItem>("Cauchy 2D");
    addItem<FTDistribution2DGaussItem>("Gauss 2D");
    addItem<FTDistribution2DGateItem>("Gate 2D");
    addItem<FTDistribution2DConeItem>("Cone 2D");
    addItem<FTDistribution2DVoigtItem>("Voigt 2D");

    addItem<FTDecayFunction1DCauchyItem>("Cauchy 1D");
    addItem<FTDecayFunction1DGaussItem>("Gauss 1D");
    addItem<FTDecayFunction1DTriangleItem>("Triangle 1D");
    addItem<FTDecayFunction1DVoigtItem>("Voigt 1D");

    addItem<FTDecayFunction2DCauchyItem>("Cauchy 2D");
    addItem<FTDecayFunction2DGaussItem>("Gauss 2D");
    addItem<FTDecayFunction2DVoigtItem>("Voigt 2D");

    addItem<BasicLattice2DItem>("Basic");
    addItem<SquareLattice2DItem>("Square");
    addItem<HexagonalLattice2DItem>("Hexagonal");

    addItem<JobItem>();

    addItem<IntensityDataItem>();
    addItem<SpecularDataItem>();
    addItem<Data1DViewItem>();
    addItem<DataPropertyContainer>();
    addItem<Data1DProperties>();

    addItem<BasicAxisItem>();
    addItem<PointwiseAxisItem>();
    addItem<AmplitudeAxisItem>();

    addItem<BeamWavelengthItem>();
    addItem<BeamAzimuthalAngleItem>();
    addItem<BeamInclinationAngleItem>();
    addItem<SpecularBeamInclinationItem>();
    addItem<SpecularBeamWavelengthItem>();

    addItem<ResolutionFunctionNoneItem>();
    addItem<ResolutionFunction2DGaussianItem>();

    addItem<MaskContainerItem>();
    addItem<RectangleItem>();
    addItem<PolygonPointItem>();
    addItem<PolygonItem>();
    addItem<VerticalLineItem>();
    addItem<HorizontalLineItem>();
    addItem<EllipseItem>();
    addItem<MaskAllItem>();

    addItem<RegionOfInterestItem>();

    addItem<GroupItem>();

    addItem<ParameterContainerItem>();
    addItem<ParameterLabelItem>();
    addItem<ParameterItem>();

    addItem<FitParameterContainerItem>();
    addItem<FitParameterItem>();
    addItem<FitParameterLinkItem>();
    addItem<FitSuiteItem>();

    addItem<RealDataItem>();

    addItem<MinimizerContainerItem>();
    addItem<MinuitMinimizerItem>();
    addItem<GSLMultiMinimizerItem>();
    addItem<GeneticMinimizerItem>();
    addItem<SimAnMinimizerItem>();
    addItem<GSLLMAMinimizerItem>();

    addItem<LimitlessItem>();
    addItem<PositiveItem>();
    addItem<NonnegativeItem>();
    addItem<LowerLimitedItem>();
    addItem<UpperLimitedItem>();
    addItem<LimitedItem>();

    addItem<ProjectionContainerItem>();

    addItem<DepthProbeInstrumentItem>();
}

ItemCatalog& ItemCatalog::instance()
{
    static ItemCatalog catalog;
    return catalog;
}

std::unique_ptr<SessionItem> ItemCatalog::createItemPtr(const QString& modelType) const
{
    if (!m_data.contains(modelType))
        throw Error("GUI::Model::ItemFactory::createItem() -> Error: Model name does not exist: "
                    + modelType);

    return m_data.createItemPtr(modelType);
}

QStringList ItemCatalog::validTopItemTypes()
{
    return {MultiLayerItem::M_TYPE,
            LayerItem::M_TYPE,
            ParticleLayoutItem::M_TYPE,
            ParticleItem::M_TYPE,
            ParticleCoreShellItem::M_TYPE,
            ParticleCompositionItem::M_TYPE,
            MesoCrystalItem::M_TYPE,
            TransformationItem::M_TYPE,
            Interference1DLatticeItem::M_TYPE,
            Interference2DLatticeItem::M_TYPE,
            Interference2DParaCrystalItem::M_TYPE,
            InterferenceFinite2DLatticeItem::M_TYPE,
            InterferenceHardDiskItem::M_TYPE,
            InterferenceRadialParaCrystalItem::M_TYPE};
}

QStringList ItemCatalog::formFactorHardParticleTypes()
{
    return {AnisoPyramidItem::M_TYPE,
            BoxItem::M_TYPE,
            ConeItem::M_TYPE,
            Cone6Item::M_TYPE,
            CuboctahedronItem::M_TYPE,
            CylinderItem::M_TYPE,
            DodecahedronItem::M_TYPE,
            EllipsoidalCylinderItem::M_TYPE,
            FullSphereItem::M_TYPE,
            FullSpheroidItem::M_TYPE,
            HemiEllipsoidItem::M_TYPE,
            IcosahedronItem::M_TYPE,
            Prism3Item::M_TYPE,
            Prism6Item::M_TYPE,
            PyramidItem::M_TYPE,
            TetrahedronItem::M_TYPE,
            TruncatedCubeItem::M_TYPE,
            TruncatedSphereItem::M_TYPE,
            TruncatedSpheroidItem::M_TYPE};
}

QStringList ItemCatalog::formFactorRippleTypes()
{
    return {BoxItem::M_TYPE,
            BarGaussItem::M_TYPE,
            BarLorentzItem::M_TYPE,
            CosineRippleBoxItem::M_TYPE,
            CosineRippleGaussItem::M_TYPE,
            CosineRippleLorentzItem::M_TYPE,
            SawtoothRippleBoxItem::M_TYPE,
            SawtoothRippleGaussItem::M_TYPE,
            SawtoothRippleLorentzItem::M_TYPE};
}

QStringList ItemCatalog::formFactorTypes()
{
    // Fill own list instead of just concatenating the above ones, because the sort order differs
    return {AnisoPyramidItem::M_TYPE,
            BarGaussItem::M_TYPE,
            BarLorentzItem::M_TYPE,
            BoxItem::M_TYPE,
            ConeItem::M_TYPE,
            Cone6Item::M_TYPE,
            CuboctahedronItem::M_TYPE,
            CylinderItem::M_TYPE,
            DodecahedronItem::M_TYPE,
            EllipsoidalCylinderItem::M_TYPE,
            FullSphereItem::M_TYPE,
            FullSpheroidItem::M_TYPE,
            HemiEllipsoidItem::M_TYPE,
            IcosahedronItem::M_TYPE,
            Prism3Item::M_TYPE,
            Prism6Item::M_TYPE,
            PyramidItem::M_TYPE,
            CosineRippleBoxItem::M_TYPE,
            CosineRippleGaussItem::M_TYPE,
            CosineRippleLorentzItem::M_TYPE,
            SawtoothRippleBoxItem::M_TYPE,
            SawtoothRippleGaussItem::M_TYPE,
            SawtoothRippleLorentzItem::M_TYPE,
            TetrahedronItem::M_TYPE,
            TruncatedCubeItem::M_TYPE,
            TruncatedSphereItem::M_TYPE,
            TruncatedSpheroidItem::M_TYPE};
}

QStringList ItemCatalog::interferenceFunctionTypes()
{
    return {InterferenceRadialParaCrystalItem::M_TYPE, Interference2DParaCrystalItem::M_TYPE,
            Interference1DLatticeItem::M_TYPE,         Interference2DLatticeItem::M_TYPE,
            InterferenceFinite2DLatticeItem::M_TYPE,   InterferenceHardDiskItem::M_TYPE};
}

QStringList ItemCatalog::particleAssemblyTypes()
{
    return {ParticleCoreShellItem::M_TYPE, ParticleCompositionItem::M_TYPE,
            MesoCrystalItem::M_TYPE};
}

bool ItemCatalog::isFormFactorModelType(const QString& modeltype)
{
    return formFactorHardParticleTypes().indexOf(modeltype) != -1
           || formFactorRippleTypes().indexOf(modeltype) != -1;
}

ItemCatalog::UiInfo ItemCatalog::uiInfo(const QString& modelType) const
{
    ASSERT(m_uiInfos.contains(modelType));
    return m_uiInfos[modelType];
}


QList<ItemCatalog::UiInfo> ItemCatalog::uiInfos(const QStringList& modelTypes) const
{
    QList<UiInfo> result;
    for (const auto& modeltype : modelTypes)
        result << uiInfo(modeltype);
    return result;
}

template <typename T> void ItemCatalog::addItem()
{
    m_data.registerItem(T::M_TYPE, [] { return new T(); });
}

template <typename T>
void ItemCatalog::addItem(const QString& menuEntry, const QString& iconPath,
                          const QString& description)
{
    addItem<T>();

    UiInfo info;
    info.menuEntry = menuEntry;
    info.iconPath = iconPath;
    info.description = description;

    m_uiInfos[T::M_TYPE] = info;
}
