//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Instrument/BeamItems.h
//! @brief     Defines BeamItem hierarchy
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_INSTRUMENT_BEAMITEMS_H
#define BORNAGAIN_GUI_MODEL_INSTRUMENT_BEAMITEMS_H

#include <heinz/Vectors3D.h>
#include "GUI/Model/Instrument/SpecularBeamInclinationItem.h"
#include "GUI/Model/Session/SessionItem.h"

class BasicAxisItem;
class Beam;
class BeamAzimuthalAngleItem;
class BeamWavelengthItem;
class FootprintItem;
class GroupItem;
class IAxis;
class VectorItem;

class BeamItem : public SessionItem {
private:
    static constexpr auto P_INTENSITY{"Intensity"};
    static constexpr auto P_WAVELENGTH{"Wavelength"};
    static constexpr auto P_INCLINATION_ANGLE{"InclinationAngle"};
    static constexpr auto P_AZIMUTHAL_ANGLE{"AzimuthalAngle"};

public:
    ~BeamItem() override;

    double intensity() const;
    void setIntensity(double value);
    SessionItem* intensityItem() const;

    double wavelength() const;
    void setWavelength(double value);
    BeamWavelengthItem* wavelengthItem() const;

    virtual double getInclinationAngle() const = 0;
    virtual void setInclinationAngle(double value);
    virtual BeamDistributionItem* inclinationAngleItem() const;

    double getAzimuthalAngle() const;
    void setAzimuthalAngle(double value);
    BeamAzimuthalAngleItem* azimuthalAngleItem() const;

    std::unique_ptr<Beam> createBeam() const;

protected:
    explicit BeamItem(const QString& beam_model);

    template <typename T> void initWavelength();
    template <typename T> void initInclinationAngle();
};

class SpecularBeamItem : public BeamItem {

public:
    static constexpr auto M_TYPE{"SpecularBeam"};

private:
    static constexpr auto P_FOOTPRINT{"Footprint"};

public:
    SpecularBeamItem();
    ~SpecularBeamItem() override;

    double getInclinationAngle() const override;
    void setInclinationAngle(double value) override;
    SpecularBeamInclinationItem* inclinationAngleItem() const override;
    GroupItem* inclinationAxisGroup() const;
    BasicAxisItem* inclinationAxis() const;

    FootprintItem* footprint() const;
    void setGaussianFootprint(double value);
    void setSquareFootprint(double value);
    SessionItem* footprintGroupItem() const;
    template <typename T> T* setFootprintType();

    void updateFileName(const QString& filename) const;
    void updateToData(const IAxis& axis, QString units) const;

private:
    void updateWavelength();
};

class GISASBeamItem : public BeamItem {
public:
    static constexpr auto M_TYPE{"GISASBeam"};

    GISASBeamItem();
    ~GISASBeamItem() override;

    double getInclinationAngle() const override;
};

template <typename T> T* SpecularBeamItem::setFootprintType()
{
    static_assert(std::is_base_of<FootprintItem, T>::value,
                  "Class must be derived from FootprintItem");

    return setGroupPropertyType<T>(P_FOOTPRINT);
}

#endif // BORNAGAIN_GUI_MODEL_INSTRUMENT_BEAMITEMS_H
