//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Instrument/InstrumentItems.cpp
//! @brief     Implement class InstrumentItem and all its children
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Instrument/InstrumentItems.h"
#include "Base/Const/Units.h"
#include "Base/Pixel/RectangularPixel.h"
#include "Core/Simulation/DepthProbeSimulation.h"
#include "Device/Beam/Beam.h"
#include "Device/Coord/CoordSystem1D.h"
#include "Device/Coord/CoordSystem2D.h"
#include "Device/Detector/RectangularDetector.h"
#include "Device/Detector/SphericalDetector.h"
#include "GUI/Model/Data/PointwiseAxisItem.h"
#include "GUI/Model/Data/RealDataItem.h"
#include "GUI/Model/Group/GroupItem.h"
#include "GUI/Model/IO/ItemFileNameUtils.h"
#include "GUI/Model/Instrument/BackgroundItems.h"
#include "GUI/Model/Instrument/BeamWavelengthItem.h"
#include "GUI/Model/Instrument/DetectorItems.h"
#include "GUI/Model/Instrument/RectangularDetectorItem.h"
#include "GUI/Model/Instrument/SphericalDetectorItem.h"
#include "GUI/Model/Job/JobItemUtils.h"
#include "GUI/Model/Types/VectorItem.h"
#include "GUI/Util/Error.h"
#include "GUI/Util/Path.h"

namespace {

const QString background_group_label = "Type";

BasicAxisItem* addAxisGroupProperty(SessionItem* parent, const QString& tag)
{
    auto* axisItem = parent->addProperty<BasicAxisItem>(tag);
    axisItem->setToolTip("Incoming alpha range [deg]");
    axisItem->titleItem()->setVisible(false);
    axisItem->binsItem()->setToolTip("Number of points in scan");
    axisItem->lowerBoundItem()->setToolTip("Starting value [deg]");
    axisItem->upperBoundItem()->setToolTip("Ending value [deg]");

    axisItem->setTitle("alpha_i");
    axisItem->setLowerBound(0.0);
    axisItem->setUpperBound(10.0);

    return axisItem;
}

GroupInfo createDetectorGroup()
{
    GroupInfo info;
    info.add(SphericalDetectorItem::M_TYPE, "Spherical detector");
    info.add(RectangularDetectorItem::M_TYPE, "Rectangular detector");
    info.setDefaultType(SphericalDetectorItem::M_TYPE);
    return info;
}

GroupInfo createBackgroundGroup()
{
    GroupInfo info;
    info.add(NoBackgroundItem::M_TYPE, "None");
    info.add(ConstantBackgroundItem::M_TYPE, "Constant background");
    info.add(PoissonNoiseBackgroundItem::M_TYPE, "Poisson noise");
    info.setDefaultType(NoBackgroundItem::M_TYPE);
    return info;
}

void setBeamDistribution(ParameterDistribution::WhichParameter which,
                         const BeamDistributionItem& item, ISimulation& simulation)
{
    auto P_par_distr = item.getParameterDistributionForName(which);
    if (P_par_distr)
        simulation.addParameterDistribution(*P_par_distr);
}

} // namespace

//  ************************************************************************************************
//  class InstrumentItem
//  ************************************************************************************************

QString InstrumentItem::id() const
{
    return getItemValue(P_IDENTIFIER).toString();
}

void InstrumentItem::setId(const QString& id)
{
    setItemValue(P_IDENTIFIER, id);
}

void InstrumentItem::setInstrumentName(const QString& instrumentName)
{
    setItemName(instrumentName);
}

QString InstrumentItem::instrumentName() const
{
    return itemName();
}

QString InstrumentItem::description() const
{
    return getItemValue(P_DESCRIPTION).toString();
}

void InstrumentItem::setDescription(const QString& description)
{
    setItemValue(P_DESCRIPTION, description);
}

BeamItem* InstrumentItem::beamItem() const
{
    return item<BeamItem>(P_BEAM);
}

BackgroundItem* InstrumentItem::backgroundItem() const
{
    return &groupItem<BackgroundItem>(P_BACKGROUND);
}

GroupItem* InstrumentItem::backgroundGroup()
{
    return item<GroupItem>(P_BACKGROUND);
}

bool InstrumentItem::alignedWith(const RealDataItem* item) const
{
    return shape() == item->shape();
}

bool InstrumentItem::withPolarizerAnalyzer() const
{
    return getItemValue(P_WITH_POLARIZER_ANALYZER).toBool();
}

void InstrumentItem::setWithPolarizerAnalyzer(bool with)
{
    setItemValue(P_WITH_POLARIZER_ANALYZER, with);
}

R3 InstrumentItem::polarization() const
{
    return item<VectorItem>(P_POLARIZATION)->getVector();
}

void InstrumentItem::setPolarization(const R3& value)
{
    item<VectorItem>(P_POLARIZATION)->setVector(value);
}

VectorItem* InstrumentItem::polarizationItem() const
{
    return item<VectorItem>(P_POLARIZATION);
}

R3 InstrumentItem::analyzerDirection() const
{
    return item<VectorItem>(P_ANALYZER_DIRECTION)->getVector();
}

void InstrumentItem::setAnalyzerDirection(const R3& direction)
{
    item<VectorItem>(P_ANALYZER_DIRECTION)->setVector(direction);
}

VectorItem* InstrumentItem::analyzerDirectionItem() const
{
    return item<VectorItem>(P_ANALYZER_DIRECTION);
}

double InstrumentItem::analyzerEfficiency() const
{
    return getItemValue(P_ANALYZER_EFFICIENCY).toDouble();
}

void InstrumentItem::setAnalyzerEfficiency(const double efficiency)
{
    setItemValue(P_ANALYZER_EFFICIENCY, efficiency);
}

double InstrumentItem::analyzerTotalTransmission() const
{
    return getItemValue(P_ANALYZER_TOTAL_TRANSMISSION).toDouble();
}

void InstrumentItem::setAnalyzerTotalTransmission(const double transmission)
{
    setItemValue(P_ANALYZER_TOTAL_TRANSMISSION, transmission);
}

InstrumentItem::InstrumentItem(const QString& modelType) : SessionItem(modelType)
{
    setItemName(modelType);
    addProperty(P_IDENTIFIER, QUuid::createUuid().toString())->setVisible(false);
    addProperty(P_DESCRIPTION, QString())->setVisible(false);
    addProperty(P_WITH_POLARIZER_ANALYZER, false);
    addProperty<VectorItem>(P_POLARIZATION);
    addProperty<VectorItem>(P_ANALYZER_DIRECTION);
    addProperty(P_ANALYZER_EFFICIENCY, 0.0)->setLimits(RealLimits::limitless());
    addProperty(P_ANALYZER_TOTAL_TRANSMISSION, 1.0);
}

void InstrumentItem::initBackgroundGroup()
{
    static const GroupInfo backgroundGroup = createBackgroundGroup();
    auto* item = addGroupProperty(P_BACKGROUND, backgroundGroup);
    item->setDisplayName(background_group_label);
    item->setToolTip("Background type");
}

template <typename T> void InstrumentItem::addBeam()
{
    addProperty<T>(P_BEAM);
}

template <typename T> T* InstrumentItem::beam() const
{
    return item<T>(P_BEAM);
}

//  ************************************************************************************************
//  class SpecularInstrumentItem
//  ************************************************************************************************

SpecularInstrumentItem::SpecularInstrumentItem() : InstrumentItem(M_TYPE)
{
    addBeam<SpecularBeamItem>();

    initBackgroundGroup();
    beam<SpecularBeamItem>()->updateFileName(
        GUI::Model::FilenameUtils::instrumentDataFileName(id()));
}

SpecularBeamItem* SpecularInstrumentItem::beamItem() const
{
    return beam<SpecularBeamItem>();
}

SpecularInstrumentItem::~SpecularInstrumentItem() = default;

std::vector<int> SpecularInstrumentItem::shape() const
{
    auto* const axis_item = beamItem()->inclinationAxis();
    return {axis_item->binCount()};
}

void SpecularInstrumentItem::updateToRealData(const RealDataItem* item)
{
    if (shape().size() != item->shape().size())
        throw Error("Error in SpecularInstrumentItem::updateToRealData: The type "
                    "of instrument is incompatible with passed data shape.");

    const auto& data = item->nativeOutputData()->axis(0);
    beamItem()->updateToData(data, item->nativeDataUnits());
}

bool SpecularInstrumentItem::alignedWith(const RealDataItem* item) const
{
    const QString native_units = item->nativeDataUnits();
    if (native_units == "nbins") {
        return beamItem()->inclinationAxis()->modelType() == BasicAxisItem::M_TYPE
               && shape() == item->shape();
    }
    auto* axis_item = dynamic_cast<PointwiseAxisItem*>(beamItem()->inclinationAxis());
    if (!axis_item)
        return false;
    if (axis_item->getUnitsLabel() != native_units)
        return false;

    const auto* instrument_axis = axis_item->axis();
    if (!instrument_axis)
        return false;

    if (!item->hasNativeData())
        return false;

    const auto& native_axis = item->nativeOutputData()->axis(0);
    return *instrument_axis == native_axis;
}

ICoordSystem* SpecularInstrumentItem::createCoordSystem() const
{
    std::unique_ptr<Instrument> instrument(new Instrument);
    instrument->setBeam(*beamItem()->createBeam());
    // \todo setPolFilters ?
    auto* axis_item = beamItem()->inclinationAxis();
    if (auto* pointwise_axis = dynamic_cast<PointwiseAxisItem*>(axis_item)) {
        if (!pointwise_axis->containsNonXMLData()) // workaround for loading project
            return nullptr;
        Axes::Coords native_units =
            GUI::Model::JobItemUtils::coordsFromName(pointwise_axis->getUnitsLabel());
        return new AngularReflectometryCoordinates(instrument->beam().wavelength(),
                                                   *pointwise_axis->axis(), native_units);
    }

    return new AngularReflectometryCoordinates(instrument->beam().wavelength(),
                                               *axis_item->createAxis(1.0), Axes::Coords::DEGREES);
}

QString SpecularInstrumentItem::defaultName() const
{
    return "Specular";
}

//  ************************************************************************************************
//  class DepthProbeInstrumentItem
//  ************************************************************************************************

DepthProbeInstrumentItem::DepthProbeInstrumentItem() : InstrumentItem(M_TYPE)
{
    addBeam<SpecularBeamItem>();

    auto* axisItem = beamItem()->inclinationAxis();
    axisItem->setLowerBound(0.0);
    axisItem->setUpperBound(1.0);
    axisItem->setBinCount(500);

    auto* axis = addProperty<BasicAxisItem>(P_Z_AXIS);
    axis->setLowerBound(-100.0);
    axis->setUpperBound(100.0);
    axis->titleItem()->setVisible(false);
    axis->binsItem()->setToolTip("Number of points in scan across sample bulk");
    axis->lowerBoundItem()->setToolTip("Starting value below sample horizont in nm");
    axis->upperBoundItem()->setToolTip("Ending value above sample horizont in nm");
}

SpecularBeamItem* DepthProbeInstrumentItem::beamItem() const
{
    return beam<SpecularBeamItem>();
}

std::vector<int> DepthProbeInstrumentItem::shape() const
{
    return std::vector<int>(); // no certain shape to avoid linking to real data
}

void DepthProbeInstrumentItem::updateToRealData(const RealDataItem*)
{
    throw std::runtime_error("DepthProbeInstrumentItem::updateToRealData()");
}

QString DepthProbeInstrumentItem::defaultName() const
{
    return "DepthProbe";
}

std::unique_ptr<DepthProbeSimulation> DepthProbeInstrumentItem::createSimulation() const
{
    std::unique_ptr<DepthProbeSimulation> simulation = std::make_unique<DepthProbeSimulation>();

    auto* const axis_item = beamItem()->inclinationAxis();

    auto axis = axis_item->createAxis(Units::deg);

    simulation->setBeamParameters(beamItem()->wavelength(), static_cast<int>(axis->size()),
                                  axis->lowerBound(), axis->upperBound());

    auto* depthAxisItem = dynamic_cast<BasicAxisItem*>(getItem(P_Z_AXIS));
    auto depthAxis = depthAxisItem->createAxis(1.0);
    simulation->setZSpan(depthAxis->size(), depthAxis->lowerBound(), depthAxis->upperBound());

    setBeamDistribution(ParameterDistribution::BeamWavelength, *beamItem()->wavelengthItem(),
                        *simulation);

    setBeamDistribution(ParameterDistribution::BeamInclinationAngle,
                        *beamItem()->inclinationAngleItem(), *simulation);

    return simulation;
}

ICoordSystem* DepthProbeInstrumentItem::createCoordSystem() const
{
    // \todo setPolFilters ?
    return createSimulation()->createCoordSystem();
}

BasicAxisItem* DepthProbeInstrumentItem::zAxis() const
{
    return item<BasicAxisItem>(P_Z_AXIS);
}

//  ************************************************************************************************
//  class Instrument2DItem
//  ************************************************************************************************

Instrument2DItem::Instrument2DItem(const QString& modelType) : InstrumentItem(modelType)
{
    addBeam<GISASBeamItem>();
    static const GroupInfo detectorGroup = createDetectorGroup();
    addGroupProperty(P_DETECTOR, detectorGroup);
    initBackgroundGroup();

    setDefaultTag(P_DETECTOR);
}

Instrument2DItem::~Instrument2DItem() = default;

DetectorItem* Instrument2DItem::detector() const
{
    return &groupItem<DetectorItem>(P_DETECTOR);
}

GroupItem* Instrument2DItem::detectorGroup()
{
    return item<GroupItem>(P_DETECTOR);
}

void Instrument2DItem::clearMasks()
{
    detector()->clearMasks();
}

void Instrument2DItem::importMasks(const MaskContainerItem* maskContainer)
{
    detector()->importMasks(maskContainer);
}

std::unique_ptr<Instrument> Instrument2DItem::createInstrument() const
{
    std::unique_ptr<Instrument> instrument(new Instrument);
    instrument->setBeam(*beamItem()->createBeam());
    instrument->setDetector(*detector()->createDetector());
    instrument->setPolFilters(polarization(), analyzerDirection(), analyzerEfficiency(),
                              analyzerTotalTransmission());

    return instrument;
}

bool Instrument2DItem::isDetectorPropertyName(const QString& name)
{
    return name == P_DETECTOR;
}

//  ************************************************************************************************
//  class GISASInstrumentItem
//  ************************************************************************************************

GISASInstrumentItem::GISASInstrumentItem() : Instrument2DItem(M_TYPE) {}

std::vector<int> GISASInstrumentItem::shape() const
{
    auto* detector_item = detector();
    return {detector_item->xSize(), detector_item->ySize()};
}

void GISASInstrumentItem::updateToRealData(const RealDataItem* item)
{
    if (!item)
        return;

    const auto data_shape = item->shape();
    if (shape().size() != data_shape.size())
        throw Error("Error in GISASInstrumentItem::updateToRealData: The type of "
                    "instrument is incompatible with passed data shape.");
    detector()->setXSize(data_shape[0]);
    detector()->setYSize(data_shape[1]);
}

QString GISASInstrumentItem::defaultName() const
{
    return "GISAS";
}

ICoordSystem* GISASInstrumentItem::createCoordSystem() const
{
    const auto instrument = createInstrument();
    instrument->initDetector();
    return instrument->createScatteringCoords();
}

//  ************************************************************************************************
//  class OffSpecularInstrumentItem
//  ************************************************************************************************

OffSpecularInstrumentItem::OffSpecularInstrumentItem() : Instrument2DItem(M_TYPE)
{
    BasicAxisItem* axis_item = addAxisGroupProperty(this, P_ALPHA_AXIS);
    auto* inclination_item = axis_item->lowerBoundItem();
    auto* beam_item = beamItem();
    beam_item->setInclinationAngle(inclination_item->value().toDouble());
    beam_item->inclinationAngleItem()->setEnabled(false);
    inclination_item->mapper()->setOnValueChange([beam_item, inclination_item]() {
        beam_item->setInclinationAngle(inclination_item->value().toDouble());
    });
}

std::vector<int> OffSpecularInstrumentItem::shape() const
{
    const int x_size = item<BasicAxisItem>(P_ALPHA_AXIS)->binCount();
    auto* detector_item = detector();
    return {x_size, detector_item->ySize()};
}

void OffSpecularInstrumentItem::updateToRealData(const RealDataItem* dataItem)
{
    if (!dataItem)
        return;

    const auto data_shape = dataItem->shape();
    if (shape().size() != data_shape.size())
        throw Error("Error in OffSpecularInstrumentItem::updateToRealData: The type of "
                    "instrument is incompatible with passed data shape.");

    item<BasicAxisItem>(P_ALPHA_AXIS)->setBinCount(data_shape[0]);
    detector()->setYSize(data_shape[1]);
}

QString OffSpecularInstrumentItem::defaultName() const
{
    return "OffSpecular";
}

ICoordSystem* OffSpecularInstrumentItem::createCoordSystem() const
{
    const auto instrument = createInstrument();
    instrument->initDetector();
    auto* const axis_item = item<BasicAxisItem>(OffSpecularInstrumentItem::P_ALPHA_AXIS);
    const auto* const detector2d = dynamic_cast<const IDetector2D*>(instrument->getDetector());
    const auto axes = detector2d->axesClippedToRegionOfInterest();
    ASSERT(axes.size() == 2);
    const IAxis& yAxis = *axes[1];
    const Beam& beam = instrument->beam();
    const auto alphaAxis = axis_item->createAxis(Units::deg);

    if (const auto* rectDetector = dynamic_cast<const RectangularDetector*>(detector2d)) {
        std::unique_ptr<RectangularPixel> detectorPixel(rectDetector->regionOfInterestPixel());
        return OffSpecularCoordinates::createForRectangularDetector(beam.direction(), *alphaAxis,
                                                                    *detectorPixel, yAxis);
    }
    if (dynamic_cast<const SphericalDetector*>(detector2d))
        return OffSpecularCoordinates::createForSphericalDetector(beam.direction(), *alphaAxis,
                                                                  yAxis);

    ASSERT(0);
}

BasicAxisItem* OffSpecularInstrumentItem::alphaAxis() const
{
    return item<BasicAxisItem>(P_ALPHA_AXIS);
}
