//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Instrument/InstrumentItems.h
//! @brief     Defines class InstrumentItem and all its children
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_INSTRUMENT_INSTRUMENTITEMS_H
#define BORNAGAIN_GUI_MODEL_INSTRUMENT_INSTRUMENTITEMS_H

#include "GUI/Model/Instrument/BeamItems.h" // inheritance prevents forward declaration

class BackgroundItem;
class DataItem;
class DepthProbeSimulation;
class DetectorItem;
class GroupItem;
class Instrument;
class ICoordSystem;
class MaskContainerItem;
class RealDataItem;

//! Abstract base class for instrument-specific item classes.

class BA_CORE_API_ InstrumentItem : public SessionItem {
private:
    static constexpr auto P_IDENTIFIER{"Identifier"};
    static constexpr auto P_BEAM{"Beam"};
    static constexpr auto P_BACKGROUND{"Background"};
    static constexpr auto P_DESCRIPTION{"Description"};
    static constexpr auto P_WITH_POLARIZER_ANALYZER{"WithPolarizerAnalyzer"};
    static constexpr auto P_POLARIZATION{"Polarization"};
    static constexpr auto P_ANALYZER_DIRECTION{"Analyzer direction"};
    static constexpr auto P_ANALYZER_EFFICIENCY{"Efficiency"};
    static constexpr auto P_ANALYZER_TOTAL_TRANSMISSION{"Transmission"};

public:
    QString id() const;
    void setId(const QString& id);

    void setInstrumentName(const QString& instrumentName);
    QString instrumentName() const;

    QString description() const;
    void setDescription(const QString& description);

    virtual BeamItem* beamItem() const;
    BackgroundItem* backgroundItem() const;
    GroupItem* backgroundGroup();
    template <typename T> T* setBackgroundType();

    virtual std::vector<int> shape() const = 0;
    virtual void clearMasks() {}
    virtual void importMasks(const MaskContainerItem*) {}
    virtual void updateToRealData(const RealDataItem* item) = 0;
    virtual bool alignedWith(const RealDataItem* item) const;

    //! The default user visible name when creating an instrument
    virtual QString defaultName() const = 0;

    virtual ICoordSystem* createCoordSystem() const = 0;

    bool withPolarizerAnalyzer() const;
    void setWithPolarizerAnalyzer(bool with);

    R3 polarization() const;
    void setPolarization(const R3& value);
    VectorItem* polarizationItem() const;

    R3 analyzerDirection() const;
    void setAnalyzerDirection(const R3& direction);
    VectorItem* analyzerDirectionItem() const;

    double analyzerEfficiency() const;
    void setAnalyzerEfficiency(double efficiency);

    double analyzerTotalTransmission() const;
    void setAnalyzerTotalTransmission(double transmission);

protected:
    explicit InstrumentItem(const QString& modelType);

    void initBackgroundGroup();

    template <typename T> void addBeam();
    template <typename T> T* beam() const;
};


class SpecularInstrumentItem : public InstrumentItem {
public:
    static constexpr auto M_TYPE{"SpecularInstrument"};

    SpecularInstrumentItem();
    ~SpecularInstrumentItem() override;

    SpecularBeamItem* beamItem() const override;

    std::vector<int> shape() const override;
    void updateToRealData(const RealDataItem* item) override;
    bool alignedWith(const RealDataItem* item) const override;
    QString defaultName() const override;

    ICoordSystem* createCoordSystem() const override;
};


class DepthProbeInstrumentItem : public InstrumentItem {
private:
    static constexpr auto P_Z_AXIS{"Z axis"};

public:
    static constexpr auto M_TYPE{"DepthProbeInstrument"};

    DepthProbeInstrumentItem();

    SpecularBeamItem* beamItem() const override;

    std::vector<int> shape() const override;
    void updateToRealData(const RealDataItem* item) override;
    QString defaultName() const override;

    // FIXME switch to base ISimulation class after InstrumentItem refactoring and
    // after ISimulation gets createCoordSystem method
    std::unique_ptr<DepthProbeSimulation> createSimulation() const;

    ICoordSystem* createCoordSystem() const override;

    BasicAxisItem* zAxis() const;
};


class Instrument2DItem : public InstrumentItem {
private:
    static constexpr auto P_DETECTOR{"Detector"};

public:
    ~Instrument2DItem() override;

    DetectorItem* detector() const;
    GroupItem* detectorGroup();
    template <typename T> T* setDetectorType();

    void clearMasks() override;
    void importMasks(const MaskContainerItem* maskContainer) override;

    std::unique_ptr<Instrument> createInstrument() const;

    static bool isDetectorPropertyName(const QString& name);

protected:
    explicit Instrument2DItem(const QString& modelType);
};


class GISASInstrumentItem : public Instrument2DItem {
public:
    static constexpr auto M_TYPE{"GISASInstrument"};

    GISASInstrumentItem();
    std::vector<int> shape() const override;
    void updateToRealData(const RealDataItem* item) override;
    QString defaultName() const override;

    ICoordSystem* createCoordSystem() const override;
};


class OffSpecularInstrumentItem : public Instrument2DItem {
private:
    static constexpr auto P_ALPHA_AXIS{"Alpha axis"};

public:
    static constexpr auto M_TYPE{"OffSpecularInstrument"};

    OffSpecularInstrumentItem();
    std::vector<int> shape() const override;
    void updateToRealData(const RealDataItem* item) override;
    QString defaultName() const override;

    ICoordSystem* createCoordSystem() const override;

    BasicAxisItem* alphaAxis() const;
};

//  ************************************************************************************************
//  templated functions implementation
//  ************************************************************************************************

template <typename T> T* InstrumentItem::setBackgroundType()
{
    static_assert(std::is_base_of<BackgroundItem, T>::value,
                  "Class must be derived from BackgroundItem");

    return setGroupPropertyType<T>(P_BACKGROUND);
}

template <typename T> T* Instrument2DItem::setDetectorType()
{
    static_assert(std::is_base_of<DetectorItem, T>::value,
                  "Class must be derived from DetectorItem");

    return setGroupPropertyType<T>(P_DETECTOR);
}

#endif // BORNAGAIN_GUI_MODEL_INSTRUMENT_INSTRUMENTITEMS_H
