//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Material/MaterialModel.cpp
//! @brief     Implements class MaterialModel
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Material/MaterialModel.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Material/MaterialItem.h"
#include "GUI/Model/Session/SessionXML.h"
#include "GUI/Util/DeserializationException.h"
#include <QColor>
#include <QUuid>
#include <QXmlStreamWriter>
#include <random>

using namespace GUI::Session::XML;


namespace {

QColor suggestMaterialColor(const QString& name)
{
    if (name.contains("Vacuum"))
        return QColor(179, 242, 255);
    if (name.contains("Substrate"))
        return QColor(205, 102, 0);
    if (name.contains("Default"))
        return QColor(Qt::green);
    if (name.contains("Particle"))
        return QColor(146, 198, 255);

    // return a random color
    static std::random_device r;
    std::default_random_engine re(r());
    std::uniform_int_distribution<int> ru(0, 255);

    return QColor(ru(re), ru(re), ru(re));
}

} // namespace


MaterialModel::MaterialModel(QObject* parent) : QObject(parent) {}

MaterialModel::~MaterialModel()
{
    clear();
}

void MaterialModel::clear()
{
    qDeleteAll(m_materialItems);
    m_materialItems.clear();
}

MaterialModel* MaterialModel::createCopy() const
{
    auto* result = new MaterialModel();

    for (const auto* m : m_materialItems)
        result->m_materialItems << new MaterialItem(*m);

    return result;
}

MaterialItem* MaterialModel::addRefractiveMaterial(const QString& name, double delta, double beta)
{
    auto* materialItem = new MaterialItem();
    materialItem->setMaterialName(name);
    materialItem->setColor(suggestMaterialColor(name));
    materialItem->setRefractiveIndex(delta, beta);
    m_materialItems << materialItem;
    emit materialAddedOrRemoved();

    return materialItem;
}

MaterialItem* MaterialModel::addSLDMaterial(const QString& name, double sld, double abs_term)
{
    auto* materialItem = new MaterialItem();
    materialItem->setMaterialName(name);
    materialItem->setColor(suggestMaterialColor(name));
    materialItem->setScatteringLengthDensity(complex_t(sld, abs_term));
    m_materialItems << materialItem;
    emit materialAddedOrRemoved();

    return materialItem;
}

MaterialItem* MaterialModel::materialFromName(const QString& name) const
{
    for (auto* materialItem : m_materialItems)
        if (materialItem->materialName() == name)
            return materialItem;

    return nullptr;
}

MaterialItem* MaterialModel::materialFromIdentifier(const QString& identifier) const
{
    for (auto* materialItem : m_materialItems)
        if (materialItem->identifier() == identifier)
            return materialItem;

    return nullptr;
}

MaterialItem* MaterialModel::insertCopy(const MaterialItem& material)
{
    auto* newMaterial = new MaterialItem(material);
    newMaterial->createNewIdentifier();
    newMaterial->setMaterialName(material.materialName() + " (copy)");
    m_materialItems << newMaterial;
    emit materialAddedOrRemoved();

    return newMaterial;
}

const QVector<MaterialItem*>& MaterialModel::materialItems() const
{
    return m_materialItems;
}

MaterialItem* MaterialModel::findMaterialItem(const QString& identifier) const
{
    for (auto* m : m_materialItems)
        if (m->identifier() == identifier)
            return m;

    return nullptr;
}

MaterialItem* MaterialModel::defaultMaterial() const
{
    ASSERT(!materialItems().isEmpty());
    return materialItems().front();
}

void MaterialModel::removeMaterial(const QString& identifier)
{
    removeMaterial(materialFromIdentifier(identifier));
}

void MaterialModel::removeMaterial(MaterialItem* materialItem)
{
    m_materialItems.removeAll(materialItem);
    delete materialItem;
    emit materialAddedOrRemoved();
}

void MaterialModel::writeContentTo(QXmlStreamWriter* writer) const
{
    writeAttribute(writer, GUI::Session::XML::Version, int(1));
    for (auto m : m_materialItems) {
        writer->writeStartElement("Material");
        m->writeContentTo(writer);
        writer->writeEndElement();
    }
}

void MaterialModel::readContentFrom(QXmlStreamReader* reader)
{
    const int version = reader->attributes().value(GUI::Session::XML::Version).toInt();

    if (version < 1)
        throw DeserializationException::tooOld();

    if (version > 1)
        throw DeserializationException::tooNew();

    while (reader->readNextStartElement()) {
        if (reader->name() == "Material") {
            MaterialItem* m = new MaterialItem;
            m->readContentFrom(reader);
            m_materialItems << m;
            reader->skipCurrentElement();
        }
    }
}

void MaterialModel::initFrom(const MaterialModel& from)
{
    // update existing to new contents (do not delete and recreate to keep references valid)
    for (auto* destItem : m_materialItems)
        if (auto* fromItem = from.findMaterialItem(destItem->identifier()))
            destItem->updateFrom(*fromItem);

    bool anyAddedOrRemoved = false;

    // remove non-existing
    auto iter = m_materialItems.begin();
    while (iter != m_materialItems.end())
        if (!from.findMaterialItem((*iter)->identifier())) {
            delete *iter;
            iter = m_materialItems.erase(iter);
            anyAddedOrRemoved = true;
        } else
            iter++;

    // copy new ones
    for (const auto* m : from.materialItems())
        if (!findMaterialItem(m->identifier())) {
            m_materialItems << new MaterialItem(*m);
            anyAddedOrRemoved = true;
        }

    // copy order
    QVector<MaterialItem*> tmp;
    for (const auto* m : from.materialItems())
        tmp << findMaterialItem(m->identifier());
    m_materialItems = tmp;

    if (anyAddedOrRemoved)
        emit materialAddedOrRemoved();
}

bool MaterialModel::operator==(const MaterialModel& other) const
{
    if (m_materialItems.size() != other.m_materialItems.size())
        return false;

    for (int i = 0; i < m_materialItems.size(); i++)
        if (*m_materialItems[i] != *other.m_materialItems[i])
            return false;

    return true;
}

bool MaterialModel::operator!=(const MaterialModel& other) const
{
    return !operator==(other);
}
