//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/LayerItem.cpp
//! @brief     Implements class LayerItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/LayerItem.h"
#include "GUI/Model/Group/GroupInfo.h"
#include "GUI/Model/Sample/LayerRoughnessItems.h"
#include "GUI/Model/Sample/MesoCrystalItem.h"
#include "GUI/Model/Sample/MultiLayerItem.h"
#include "GUI/Model/Sample/ParticleCompositionItem.h"
#include "GUI/Model/Sample/ParticleCoreShellItem.h"
#include "GUI/Model/Sample/ParticleItem.h"
#include "GUI/Model/Sample/ParticleLayoutItem.h"
#include "GUI/Model/Types/DoubleDescriptor.h"
#include "GUI/Model/Types/UIntDescriptor.h"
#include "GUI/Util/Error.h"
#include <QColor>
#include <boost/polymorphic_cast.hpp>

using boost::polymorphic_cast;

namespace {

const QString layer_nslices_tooltip = "Number of horizontal slices.\n"
                                      "Used for Average Layer Material calculations \n"
                                      "when corresponding simulation option is set.";

QVector<ItemWithMaterial*> layoutItemsWithMaterial(ParticleLayoutItem* item)
{
    QVector<ItemWithMaterial*> ret;

    QVector<ItemWithParticles*> itemsWithParticles{item->particles()};
    while (!itemsWithParticles.empty()) {
        auto* item = itemsWithParticles.takeFirst();
        if (!item)
            continue;

        if (item->hasModelType<ParticleCompositionItem>())
            itemsWithParticles << polymorphic_cast<ParticleCompositionItem*>(item)->particles();
        else if (item->hasModelType<MesoCrystalItem>())
            itemsWithParticles << polymorphic_cast<MesoCrystalItem*>(item)->basisParticle();
        else if (item->hasModelType<ParticleItem>())
            ret << dynamic_cast<ParticleItem*>(item);
        else if (item->hasModelType<ParticleCoreShellItem>()) {
            if (auto* core = dynamic_cast<ParticleCoreShellItem*>(item)->core())
                ret << core;
            if (auto* shell = dynamic_cast<ParticleCoreShellItem*>(item)->shell())
                ret << shell;
        } else
            throw Error("Error in GUI::MaterialUtil::materialProperties: cannot handle "
                        "passed model type '"
                        + item->modelType() + "'");
    }
    return ret;
}

} // namespace

LayerItem::LayerItem() : SessionItem(M_TYPE), ItemWithMaterial(M_TYPE)
{
    setToolTip("A layer with thickness and material");
    addProperty(P_THICKNESS, 0.0)
        ->setLimits(RealLimits::lowerLimited(0.0))
        .setToolTip("Thickness of a layer in nanometers");

    addProperty(P_NSLICES, 1)
        ->setLimits(RealLimits::lowerLimited(0.0))
        .setToolTip(layer_nslices_tooltip);

    addProperty(P_COLOR, QColor()); // initially no color

    GroupInfo info;
    info.add(LayerBasicRoughnessItem::M_TYPE, "Basic");
    info.add(LayerZeroRoughnessItem::M_TYPE, "No");
    info.setDefaultType(LayerZeroRoughnessItem::M_TYPE);
    addGroupProperty(P_ROUGHNESS, info)->setToolTip("Roughness of top interface");

    registerTag(T_LAYOUTS, 0, -1, {ParticleLayoutItem::M_TYPE});
    setDefaultTag(T_LAYOUTS);

    mapper()->setOnParentChange([this](SessionItem* new_parent) { updateAppearance(new_parent); });
}

QVector<ItemWithMaterial*> LayerItem::itemsWithMaterial()
{
    QVector<ItemWithMaterial*> result;
    result.push_back(this);
    for (auto* layout : layouts())
        result.append(layoutItemsWithMaterial(layout));
    return result;
}

SessionItem* LayerItem::thicknessItem() const
{
    return getItem(P_THICKNESS);
}

DoubleDescriptor LayerItem::thickness() const
{
    DoubleDescriptor d(getItem(P_THICKNESS), Unit::nanometer);
    d.tooltip = "Thickness of the layer";
    return d;
}

bool LayerItem::isThicknessPropertyName(const QString& name)
{
    return name == P_THICKNESS;
}

void LayerItem::setThicknessEnabled(bool enabled)
{
    getItem(P_THICKNESS)->setEnabled(enabled);
}

SelectionDescriptor<std::variant<LayerZeroRoughnessItem*, LayerBasicRoughnessItem*>>
LayerItem::roughness()
{
    SelectionDescriptor<std::variant<LayerZeroRoughnessItem*, LayerBasicRoughnessItem*>> d(
        item<GroupItem>(P_ROUGHNESS));

    d.currentItem = [=]() -> std::variant<LayerZeroRoughnessItem*, LayerBasicRoughnessItem*> {
        if (auto* r = dynamic_cast<LayerBasicRoughnessItem*>(getGroupItem(P_ROUGHNESS)))
            return {r};
        return {dynamic_cast<LayerZeroRoughnessItem*>(getGroupItem(P_ROUGHNESS))};
    };

    return d;
}

SessionItem* LayerItem::roughnessItem() const
{
    return getItem(P_ROUGHNESS);
}

void LayerItem::setRoughnessEnabled(bool enabled)
{
    getItem(P_ROUGHNESS)->setEnabled(enabled);
}

UIntDescriptor LayerItem::numSlices() const
{
    return UIntDescriptor(getItem(P_NSLICES), Unit::unitless);
}

QVector<ParticleLayoutItem*> LayerItem::layouts() const
{
    return items<ParticleLayoutItem>(T_LAYOUTS);
}

void LayerItem::removeLayout(ParticleLayoutItem* layout)
{
    model()->removeItem(layout);
}

QColor LayerItem::color() const
{
    return getItemValue(P_COLOR).value<QColor>();
}

void LayerItem::setColor(const QColor& color)
{
    setItemValue(P_COLOR, color);
}

void LayerItem::updateAppearance(SessionItem* new_parent)
{
    if (!new_parent) {
        if (parent() && parent()->hasModelType<MultiLayerItem>()) {
            // we are about to be removed from MultiLayer
            getItem(LayerItem::P_ROUGHNESS)->setEnabled(true);
            getItem(LayerItem::P_THICKNESS)->setEnabled(true);
        }
    }
}
