//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/DistributionForms.cpp
//! @brief     Implements class InstrumentSelectorWidget
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/DistributionForms.h"
#include "GUI/Model/Group/DistributionItems.h"
#include "GUI/Model/Instrument/ItemWithDistribution.h"
#include "GUI/View/Common/ScientificSpinBox.h"
#include "GUI/View/Instrument/DistributionEditor.h"

#include <QDoubleSpinBox>
#include <QFormLayout>
#include <QSpinBox>

#include <boost/polymorphic_cast.hpp>
using boost::polymorphic_downcast;

#include <boost/numeric/interval.hpp>
using boost::numeric::interval;

namespace {

namespace Range {

// constant for often used interval ]-inf,inf[
static const interval<double> full(std::numeric_limits<double>::lowest(),
                                   std::numeric_limits<double>::max());

// constant for often used interval [0,inf[
static const interval<double> nonnegative(0, std::numeric_limits<double>::max());
} // namespace Range

/// perform a strandard configuration for the QSpinBox
void configSpinBox(QSpinBox* spinbox)
{
    spinbox->setRange(1, std::numeric_limits<int>::max());
}

/// configure the given QDoubleSpinBox
void configSpinBox(QDoubleSpinBox* spinbox, const interval<double>& range, int decimals = 3)
{
    spinbox->setRange(range.lower(), range.upper());
    spinbox->setDecimals(decimals);
    spinbox->setSingleStep(0.01);
}

/// configure the given ScientificSpinBox
void configSpinBox(ScientificSpinBox* spinbox, const interval<double>& range, int decimals)
{
    spinbox->setMinimum(range.lower());
    spinbox->setMaximum(range.upper());
    spinbox->setDecimals(decimals);
    spinbox->setSingleStep(0.01);
}

/// visitor to set the value of the spin box
struct set_value_visitor {
    double value;

    set_value_visitor(double val) : value(val) {}
    void operator()(QDoubleSpinBox* spinbox) const
    {
        if (spinbox)
            spinbox->setValue(value);
    }
    void operator()(ScientificSpinBox* spinbox) const { spinbox->setValue(value); }
};

/// visitor to get the widget from the variant
struct widget_visitor {
    QWidget* operator()(QDoubleSpinBox* spinbox) { return spinbox; }
    QWidget* operator()(ScientificSpinBox* spinbox) { return spinbox; }
};

/// visitor to configure the actual spinbox
struct config_visitor {
    const interval<double>& range;
    int decimals;

    config_visitor(const interval<double>& r, int d) : range(r), decimals(d) {}
    void operator()(QDoubleSpinBox* spinbox) { configSpinBox(spinbox, range, decimals); }
    void operator()(ScientificSpinBox* spinbox) { configSpinBox(spinbox, range, decimals); }
};

} // namespace

//==================================================================================================
// DistributionForm
//==================================================================================================
//--------------------------------------------------------------------------------------------------
// protected member functions
//--------------------------------------------------------------------------------------------------

DistributionForm::DistributionForm(QWidget* parent, Qt::WindowFlags f) : QWidget(parent, f) {}

//==================================================================================================
// CosineDistributionForm
//==================================================================================================
//--------------------------------------------------------------------------------------------------
// public member functions
//--------------------------------------------------------------------------------------------------

CosineDistributionForm::CosineDistributionForm(const std::optional<MeanConfig>& mean_config,
                                               QWidget* parent, Qt::WindowFlags f)
    : DistributionForm(parent, f), m_item(nullptr)
{
    auto* layout = new QFormLayout(this);
    layout->setMargin(0);

    if (mean_config) {
        if (mean_config->scientific) {
            auto* spinbox = new ScientificSpinBox(this);
            connect(spinbox, &ScientificSpinBox::valueChanged, this,
                    &CosineDistributionForm::onMeanChanged);
            m_meanSpinBox = spinbox;
        } else {
            auto* spinbox = new QDoubleSpinBox(this);
            connect(spinbox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
                    &CosineDistributionForm::onMeanChanged);
            m_meanSpinBox = spinbox;
        }
        std::visit(config_visitor(mean_config->range, mean_config->decimals), m_meanSpinBox);
        layout->addRow(u8"Mean (\u03bc):", std::visit(widget_visitor(), m_meanSpinBox));
    } else
        m_meanSpinBox.emplace<0>(nullptr);

    m_sigmaSpinBox = new QDoubleSpinBox(this);
    configSpinBox(m_sigmaSpinBox, Range::nonnegative);
    connect(m_sigmaSpinBox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
            &CosineDistributionForm::onSigmaChanged);
    layout->addRow(u8"Standard deviation (\u03c3):", m_sigmaSpinBox);

    m_samplesSpinBox = new QSpinBox(this);
    configSpinBox(m_samplesSpinBox);
    connect(m_samplesSpinBox, qOverload<int>(&QSpinBox::valueChanged), this,
            &CosineDistributionForm::onSamplesChanged);
    layout->addRow("Number of samples:", m_samplesSpinBox);

    m_sigmaFactorSpinBox = new QDoubleSpinBox(this);
    configSpinBox(m_sigmaFactorSpinBox, Range::nonnegative);
    connect(m_sigmaFactorSpinBox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
            &CosineDistributionForm::onSigmaFactorChanged);
    layout->addRow(u8"\u03c3 factor:", m_sigmaFactorSpinBox);
}

void CosineDistributionForm::setupDistribution(ItemWithDistribution* item)
{
    if (!item->distribution()->hasModelType<DistributionCosineItem>())
        item->setDistributionType(DistributionCosineItem::M_TYPE);
    m_item = polymorphic_downcast<DistributionCosineItem*>(item->distribution());

    std::visit(set_value_visitor(m_item->mean()), m_meanSpinBox);
    m_sigmaSpinBox->setValue(m_item->sigma());
    m_samplesSpinBox->setValue(static_cast<int>(m_item->numberOfSamples()));
    m_samplesSpinBox->setMinimum(1);
    m_sigmaFactorSpinBox->setValue(m_item->sigmaFactor());
}

//--------------------------------------------------------------------------------------------------
// private slots
//--------------------------------------------------------------------------------------------------

void CosineDistributionForm::onMeanChanged(double value)
{
    m_item->setMean(value);
    emit distributionChanged();
}

void CosineDistributionForm::onSigmaChanged(double value)
{
    m_item->setSigma(value);
    emit distributionChanged();
}

void CosineDistributionForm::onSamplesChanged(int value)
{
    auto samples = static_cast<size_t>(value);
    m_item->setNumberOfSamples(samples);
    emit distributionChanged();
}

void CosineDistributionForm::onSigmaFactorChanged(double value)
{
    m_item->setSigmaFactor(value);
    emit distributionChanged();
}

//==================================================================================================
// GateDistributionForm
//==================================================================================================
//--------------------------------------------------------------------------------------------------
// public member functions
//--------------------------------------------------------------------------------------------------

GateDistributionForm::GateDistributionForm(QWidget* parent, Qt::WindowFlags f)
    : DistributionForm(parent, f), m_item(nullptr)
{
    auto* layout = new QFormLayout(this);
    layout->setMargin(0);

    m_minimumSpinBox = new QDoubleSpinBox(this);
    configSpinBox(m_minimumSpinBox, Range::full);
    connect(m_minimumSpinBox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
            &GateDistributionForm::onMinimumChanged);
    layout->addRow("Minimum:", m_minimumSpinBox);

    m_maximumSpinBox = new QDoubleSpinBox(this);
    configSpinBox(m_maximumSpinBox, Range::full);
    connect(m_maximumSpinBox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
            &GateDistributionForm::onMaximumChanged);
    layout->addRow("Maximum:", m_maximumSpinBox);

    m_samplesSpinBox = new QSpinBox(this);
    configSpinBox(m_samplesSpinBox);
    connect(m_samplesSpinBox, qOverload<int>(&QSpinBox::valueChanged), this,
            &GateDistributionForm::onSamplesChanged);
    layout->addRow("Number of samples:", m_samplesSpinBox);
}

void GateDistributionForm::setupDistribution(ItemWithDistribution* item)
{
    if (!item->distribution()->hasModelType<DistributionGateItem>())
        item->setDistributionType(DistributionGateItem::M_TYPE);
    m_item = polymorphic_downcast<DistributionGateItem*>(item->distribution());

    m_minimumSpinBox->setValue(m_item->minimum());
    m_maximumSpinBox->setValue(m_item->maximum());
    m_samplesSpinBox->setValue(static_cast<int>(m_item->numberOfSamples()));
}

//--------------------------------------------------------------------------------------------------
// private slots
//--------------------------------------------------------------------------------------------------

void GateDistributionForm::onMinimumChanged(double value)
{
    double maximum = m_item->maximum();
    if (maximum < value) {
        m_item->setRange(value, value);
        m_maximumSpinBox->setValue(value);
    } else {
        m_item->setRange(value, maximum);
        emit distributionChanged();
    }
}

void GateDistributionForm::onMaximumChanged(double value)
{
    double minimum = m_item->minimum();
    if (minimum > value) {
        m_item->setRange(value, value);
        m_minimumSpinBox->setValue(value);
    } else {
        m_item->setRange(minimum, value);
        emit distributionChanged();
    }
}

void GateDistributionForm::onSamplesChanged(int value)
{
    auto samples = static_cast<size_t>(value);
    m_item->setNumberOfSamples(samples);
    emit distributionChanged();
}

//==================================================================================================
// GaussDistributionForm
//==================================================================================================
//--------------------------------------------------------------------------------------------------
// public member functions
//--------------------------------------------------------------------------------------------------

GaussDistributionForm::GaussDistributionForm(const std::optional<MeanConfig>& mean_config,
                                             QWidget* parent, Qt::WindowFlags f)
    : DistributionForm(parent, f), m_item(nullptr)
{
    auto* layout = new QFormLayout(this);
    layout->setMargin(0);

    if (mean_config) {
        if (mean_config->scientific) {
            auto* spinbox = new ScientificSpinBox(this);
            connect(spinbox, &ScientificSpinBox::valueChanged, this,
                    &GaussDistributionForm::onMeanChanged);
            m_meanSpinBox = spinbox;
        } else {
            auto* spinbox = new QDoubleSpinBox(this);
            connect(spinbox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
                    &GaussDistributionForm::onMeanChanged);

            m_meanSpinBox = spinbox;
        }
        std::visit(config_visitor(mean_config->range, mean_config->decimals), m_meanSpinBox);
        layout->addRow(u8"Mean (\u03bc):", std::visit(widget_visitor(), m_meanSpinBox));
    } else
        m_meanSpinBox.emplace<0>(nullptr);

    m_stdDevSpinBox = new QDoubleSpinBox(this);
    configSpinBox(m_stdDevSpinBox, Range::nonnegative);
    connect(m_stdDevSpinBox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
            &GaussDistributionForm::onDeviationChanged);
    layout->addRow(u8"Standard deviation (\u03c3):", m_stdDevSpinBox);

    m_samplesSpinBox = new QSpinBox(this);
    configSpinBox(m_samplesSpinBox);
    connect(m_samplesSpinBox, qOverload<int>(&QSpinBox::valueChanged), this,
            &GaussDistributionForm::onSamplesChanged);
    layout->addRow("Number of samples:", m_samplesSpinBox);

    m_sigmaFactorSpinBox = new QDoubleSpinBox(this);
    configSpinBox(m_sigmaFactorSpinBox, Range::nonnegative);
    connect(m_sigmaFactorSpinBox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
            &GaussDistributionForm::onSigmaFactorChanged);
    layout->addRow(u8"\u03c3 factor:", m_sigmaFactorSpinBox);
}

void GaussDistributionForm::setupDistribution(ItemWithDistribution* item)
{
    if (!item->distribution()->hasModelType<DistributionGaussianItem>())
        item->setDistributionType(DistributionGaussianItem::M_TYPE);
    m_item = polymorphic_downcast<DistributionGaussianItem*>(item->distribution());

    std::visit(set_value_visitor(m_item->mean()), m_meanSpinBox);
    m_stdDevSpinBox->setValue(m_item->standardDeviation());
    m_samplesSpinBox->setValue(static_cast<int>(m_item->numberOfSamples()));
    m_sigmaFactorSpinBox->setValue(m_item->sigmaFactor());
}

//--------------------------------------------------------------------------------------------------
// private slots
//--------------------------------------------------------------------------------------------------

void GaussDistributionForm::onMeanChanged(double value)
{
    m_item->setMean(value);
    emit distributionChanged();
}

void GaussDistributionForm::onDeviationChanged(double value)
{
    m_item->setStandardDeviation(value);
    emit distributionChanged();
}

void GaussDistributionForm::onSamplesChanged(int value)
{
    auto samples = static_cast<size_t>(value);
    m_item->setNumberOfSamples(samples);
    emit distributionChanged();
}

void GaussDistributionForm::onSigmaFactorChanged(double value)
{
    m_item->setSigmaFactor(value);
    emit distributionChanged();
}

//==================================================================================================
// LogNormalDistributionForm
//==================================================================================================
//--------------------------------------------------------------------------------------------------
// public member functions
//--------------------------------------------------------------------------------------------------

LogNormalDistributionForm::LogNormalDistributionForm(QWidget* parent, Qt::WindowFlags f)
    : DistributionForm(parent, f), m_item(nullptr)
{
    auto* layout = new QFormLayout(this);
    layout->setMargin(0);

    m_medianSpinBox = new QDoubleSpinBox(this);
    configSpinBox(m_medianSpinBox, Range::nonnegative);
    connect(m_medianSpinBox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
            &LogNormalDistributionForm::onMedianChanged);
    layout->addRow("Median:", m_medianSpinBox);

    m_scaleSpinBox = new QDoubleSpinBox(this);
    configSpinBox(m_scaleSpinBox, Range::nonnegative);
    connect(m_scaleSpinBox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
            &LogNormalDistributionForm::onScaleChanged);
    layout->addRow("Scale parameter:", m_scaleSpinBox);

    m_samplesSpinBox = new QSpinBox(this);
    configSpinBox(m_samplesSpinBox);
    connect(m_samplesSpinBox, qOverload<int>(&QSpinBox::valueChanged), this,
            &LogNormalDistributionForm::onSamplesChanged);
    layout->addRow("Number of samples:", m_samplesSpinBox);

    m_sigmaFactorSpinBox = new QDoubleSpinBox(this);
    configSpinBox(m_sigmaFactorSpinBox, Range::nonnegative);
    connect(m_sigmaFactorSpinBox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
            &LogNormalDistributionForm::onSigmaFactorChanged);
    layout->addRow(u8"\u03c3 factor:", m_sigmaFactorSpinBox);
}

void LogNormalDistributionForm::setupDistribution(ItemWithDistribution* item)
{
    if (!item->distribution()->hasModelType<DistributionLogNormalItem>())
        item->setDistributionType(DistributionLogNormalItem::M_TYPE);
    m_item = polymorphic_downcast<DistributionLogNormalItem*>(item->distribution());

    m_medianSpinBox->setValue(m_item->median());
    m_scaleSpinBox->setValue(m_item->scaleParameter());
    m_samplesSpinBox->setValue(static_cast<int>(m_item->numberOfSamples()));
    m_sigmaFactorSpinBox->setValue(m_item->sigmaFactor());
}

//--------------------------------------------------------------------------------------------------
// private slots
//--------------------------------------------------------------------------------------------------

void LogNormalDistributionForm::onMedianChanged(double value)
{
    m_item->setMedian(value);
    emit distributionChanged();
}

void LogNormalDistributionForm::onScaleChanged(double value)
{
    m_item->setScaleParameter(value);
    emit distributionChanged();
}

void LogNormalDistributionForm::onSamplesChanged(int value)
{
    auto samples = static_cast<size_t>(value);
    m_item->setNumberOfSamples(samples);
    emit distributionChanged();
}

void LogNormalDistributionForm::onSigmaFactorChanged(double value)
{
    m_item->setSigmaFactor(value);
    emit distributionChanged();
}

//==================================================================================================
// LorentzDistributionForm
//==================================================================================================
//--------------------------------------------------------------------------------------------------
// public member functions
//--------------------------------------------------------------------------------------------------

LorentzDistributionForm::LorentzDistributionForm(const std::optional<MeanConfig>& mean_config,
                                                 QWidget* parent, Qt::WindowFlags f)
    : DistributionForm(parent, f), m_item(nullptr)
{
    auto* layout = new QFormLayout(this);
    layout->setMargin(0);

    if (mean_config) {
        if (mean_config->scientific) {
            auto* spinbox = new ScientificSpinBox(this);
            connect(spinbox, &ScientificSpinBox::valueChanged, this,
                    &LorentzDistributionForm::onMeanChanged);
            m_meanSpinBox = spinbox;
        } else {
            auto* spinbox = new QDoubleSpinBox(this);
            connect(spinbox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
                    &LorentzDistributionForm::onMeanChanged);
            m_meanSpinBox = spinbox;
        }
        std::visit(config_visitor(mean_config->range, mean_config->decimals), m_meanSpinBox);
        layout->addRow(u8"Mean (\u03bc):", std::visit(widget_visitor(), m_meanSpinBox));
    } else
        m_meanSpinBox.emplace<0>(nullptr);

    m_hwhmSpinBox = new QDoubleSpinBox(this);
    configSpinBox(m_hwhmSpinBox, Range::nonnegative);
    connect(m_hwhmSpinBox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
            &LorentzDistributionForm::onHWHMChanged);
    layout->addRow("HWHM:", m_hwhmSpinBox);

    m_samplesSpinBox = new QSpinBox(this);
    configSpinBox(m_samplesSpinBox);
    connect(m_samplesSpinBox, qOverload<int>(&QSpinBox::valueChanged), this,
            &LorentzDistributionForm::onSamplesChanged);
    layout->addRow("Number of samples:", m_samplesSpinBox);

    m_sigmaFactorSpinBox = new QDoubleSpinBox(this);
    configSpinBox(m_sigmaFactorSpinBox, Range::nonnegative);
    connect(m_sigmaFactorSpinBox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
            &LorentzDistributionForm::onSigmaFactorChanged);
    layout->addRow(u8"\u03c3 factor:", m_sigmaFactorSpinBox);
}

void LorentzDistributionForm::setupDistribution(ItemWithDistribution* item)
{
    if (!item->distribution()->hasModelType<DistributionLorentzItem>())
        item->setDistributionType(DistributionLorentzItem::M_TYPE);
    m_item = polymorphic_downcast<DistributionLorentzItem*>(item->distribution());

    std::visit(set_value_visitor(m_item->mean()), m_meanSpinBox);
    m_hwhmSpinBox->setValue(m_item->hwhm());
    m_samplesSpinBox->setValue(static_cast<int>(m_item->numberOfSamples()));
    m_sigmaFactorSpinBox->setValue(m_item->sigmaFactor());
}

//--------------------------------------------------------------------------------------------------
// private slots
//--------------------------------------------------------------------------------------------------

void LorentzDistributionForm::onMeanChanged(double value)
{
    m_item->setMean(value);
    emit distributionChanged();
}

void LorentzDistributionForm::onHWHMChanged(double value)
{
    m_item->setHWHM(value);
    emit distributionChanged();
}

void LorentzDistributionForm::onSamplesChanged(int value)
{
    auto samples = static_cast<size_t>(value);
    m_item->setNumberOfSamples(samples);
    emit distributionChanged();
}

void LorentzDistributionForm::onSigmaFactorChanged(double value)
{
    m_item->setSigmaFactor(value);
    emit distributionChanged();
}

//==================================================================================================
// NoneDistributionForm
//==================================================================================================
//--------------------------------------------------------------------------------------------------
// public member functions
//--------------------------------------------------------------------------------------------------

NoneDistributionForm::NoneDistributionForm(const std::optional<MeanConfig>& mean_config,
                                           QWidget* parent, Qt::WindowFlags f)
    : DistributionForm(parent, f), m_item(nullptr)
{
    auto* layout = new QFormLayout(this);
    layout->setMargin(0);

    if (mean_config) {
        if (mean_config->scientific) {
            auto* spinbox = new ScientificSpinBox(this);
            connect(spinbox, &ScientificSpinBox::valueChanged, this,
                    &NoneDistributionForm::onValueChanged);
            m_spinBox = spinbox;
        } else {
            auto* spinbox = new QDoubleSpinBox(this);
            connect(spinbox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
                    &NoneDistributionForm::onValueChanged);
            m_spinBox = spinbox;
        }
        std::visit(config_visitor(mean_config->range, mean_config->decimals), m_spinBox);
        layout->addRow("Value:", std::visit(widget_visitor(), m_spinBox));
    } else
        m_spinBox.emplace<0>(nullptr);
}

void NoneDistributionForm::setupDistribution(ItemWithDistribution* item)
{
    if (!item->distribution()->hasModelType<DistributionNoneItem>())
        item->setDistributionType(DistributionNoneItem::M_TYPE);
    m_item = polymorphic_downcast<DistributionNoneItem*>(item->distribution());

    std::visit(set_value_visitor(m_item->mean()), m_spinBox);
}

//--------------------------------------------------------------------------------------------------
// private slots
//--------------------------------------------------------------------------------------------------

void NoneDistributionForm::onValueChanged(double value)
{
    m_item->setMean(value);
    emit distributionChanged();
}

//==================================================================================================
// TrapezoidDistributionForm
//==================================================================================================
//--------------------------------------------------------------------------------------------------
// public member functions
//--------------------------------------------------------------------------------------------------

TrapezoidDistributionForm::TrapezoidDistributionForm(QWidget* parent, Qt::WindowFlags f)
    : DistributionForm(parent, f), m_item(nullptr)
{
    auto* layout = new QFormLayout(this);
    layout->setMargin(0);

    m_centerSpinBox = new QDoubleSpinBox(this);
    configSpinBox(m_centerSpinBox, Range::full);
    connect(m_centerSpinBox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
            &TrapezoidDistributionForm::onCenterChanged);
    layout->addRow("Center:", m_centerSpinBox);

    m_leftWidthSpinBox = new QDoubleSpinBox(this);
    configSpinBox(m_leftWidthSpinBox, Range::nonnegative);
    m_leftWidthSpinBox->setDecimals(3);
    m_leftWidthSpinBox->setSingleStep(0.01);
    connect(m_leftWidthSpinBox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
            &TrapezoidDistributionForm::onLeftWidthChanged);
    layout->addRow("Left width:", m_leftWidthSpinBox);

    m_middleWidthSpinBox = new QDoubleSpinBox(this);
    configSpinBox(m_middleWidthSpinBox, Range::nonnegative);
    connect(m_middleWidthSpinBox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
            &TrapezoidDistributionForm::onMiddleWidthChanged);
    layout->addRow("Middle width:", m_middleWidthSpinBox);

    m_rightWidthSpinBox = new QDoubleSpinBox(this);
    configSpinBox(m_rightWidthSpinBox, Range::nonnegative);
    connect(m_rightWidthSpinBox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
            &TrapezoidDistributionForm::onRightWidthChanged);
    layout->addRow("Right width:", m_rightWidthSpinBox);

    m_samplesSpinBox = new QSpinBox(this);
    configSpinBox(m_samplesSpinBox);
    connect(m_samplesSpinBox, qOverload<int>(&QSpinBox::valueChanged), this,
            &TrapezoidDistributionForm::onSamplesChanged);
    layout->addRow("Number of samples:", m_samplesSpinBox);
}

void TrapezoidDistributionForm::setupDistribution(ItemWithDistribution* item)
{
    if (!item->distribution()->hasModelType<DistributionTrapezoidItem>())
        item->setDistributionType(DistributionTrapezoidItem::M_TYPE);
    m_item = polymorphic_downcast<DistributionTrapezoidItem*>(item->distribution());

    m_centerSpinBox->setValue(m_item->center());
    m_leftWidthSpinBox->setValue(m_item->leftWidth());
    m_middleWidthSpinBox->setValue(m_item->middleWidth());
    m_rightWidthSpinBox->setValue(m_item->rightWidth());
    m_samplesSpinBox->setValue(static_cast<int>(m_item->numberOfSamples()));
}

//--------------------------------------------------------------------------------------------------
// private slots
//--------------------------------------------------------------------------------------------------

void TrapezoidDistributionForm::onCenterChanged(double value)
{
    m_item->setCenter(value);
    emit distributionChanged();
}

void TrapezoidDistributionForm::onLeftWidthChanged(double value)
{
    m_item->setLeftWidth(value);
    emit distributionChanged();
}

void TrapezoidDistributionForm::onMiddleWidthChanged(double value)
{
    m_item->setMiddleWidth(value);
    emit distributionChanged();
}

void TrapezoidDistributionForm::onRightWidthChanged(double value)
{
    m_item->setRightWidth(value);
    emit distributionChanged();
}

void TrapezoidDistributionForm::onSamplesChanged(int value)
{
    auto samples = static_cast<size_t>(value);
    m_item->setNumberOfSamples(samples);
    emit distributionChanged();
}
