//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/InstrumentEditor.cpp
//! @brief     Implements class InstrumentEditor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/InstrumentEditor.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Instrument/InstrumentItems.h"
#include "GUI/View/Instrument/DepthProbeInstrumentEditor.h"
#include "GUI/View/Instrument/GISASInstrumentEditor.h"
#include "GUI/View/Instrument/OffSpecularInstrumentEditor.h"
#include "GUI/View/Instrument/SpecularInstrumentEditor.h"

#include <QFormLayout>
#include <QLabel>
#include <QLineEdit>
#include <QScrollArea>
#include <QStackedLayout>

namespace {

/// generate a type name for an instrument
QString instrumentTypeName(InstrumentItem* instrument)
{
    if (instrument->hasModelType<SpecularInstrumentItem>())
        return "Specular";
    if (instrument->hasModelType<DepthProbeInstrumentItem>())
        return "Depth probe";
    if (instrument->hasModelType<GISASInstrumentItem>())
        return "GISAS";
    if (instrument->hasModelType<OffSpecularInstrumentItem>())
        return "Off specular";
    ASSERT(0);
}

} // namespace

InstrumentEditor::InstrumentEditor(QWidget* parent, Qt::WindowFlags f)
    : QWidget(parent, f), m_item(nullptr)
{
    auto* layout = new QVBoxLayout(this);
    layout->setContentsMargins(0, 20, 0, 20);

    auto* form = new QFormLayout();
    form->setMargin(17);
    layout->addLayout(form);

    m_nameLineEdit = new QLineEdit(this);
    form->addRow("Instrument name:", m_nameLineEdit);

    m_typeLabel = new QLabel(this);
    form->addRow("Instrument type:", m_typeLabel);

    m_stack = new QStackedLayout();
    layout->addLayout(m_stack);

    m_depthProbeWidget = new QScrollArea(this);
    m_depthProbeWidget->setWidgetResizable(true);
    m_depthProbeEditor = new DepthProbeInstrumentEditor(m_depthProbeWidget);
    m_depthProbeWidget->setWidget(m_depthProbeEditor);
    m_stack->addWidget(m_depthProbeWidget);

    m_gisasWidget = new QScrollArea(this);
    m_gisasWidget->setWidgetResizable(true);
    m_gisasEditor = new GISASInstrumentEditor(m_gisasWidget);
    m_gisasWidget->setWidget(m_gisasEditor);
    m_stack->addWidget(m_gisasWidget);

    m_offSpecularWidget = new QScrollArea(this);
    m_offSpecularWidget->setWidgetResizable(true);
    m_offSpecularEditor = new OffSpecularInstrumentEditor(m_offSpecularWidget);
    m_offSpecularWidget->setWidget(m_offSpecularEditor);
    m_stack->addWidget(m_offSpecularWidget);

    m_specularWidget = new QScrollArea(this);
    m_specularWidget->setWidgetResizable(true);
    m_specularEditor = new SpecularInstrumentEditor(m_specularWidget);
    m_specularWidget->setWidget(m_specularEditor);
    m_stack->addWidget(m_specularWidget);

    m_blankEditor = new QWidget(this);
    m_stack->addWidget(m_blankEditor);

    connect(m_nameLineEdit, &QLineEdit::editingFinished, this,
            &InstrumentEditor::onInstrumentNameChanged);
}

QSize InstrumentEditor::sizeHint() const
{
    return QSize(600, 600);
}

void InstrumentEditor::setItem(InstrumentItem* item)
{
    m_item = item;
    setEnabled(m_item);
    if (m_item) {
        m_nameLineEdit->setEnabled(true);
        m_nameLineEdit->setText(m_item->itemName());
        m_typeLabel->setText(instrumentTypeName(m_item));
        if (m_item->hasModelType<SpecularInstrumentItem>()) {
            m_specularEditor->setItem(dynamic_cast<SpecularInstrumentItem*>(m_item));
            m_stack->setCurrentWidget(m_specularWidget);
        } else if (m_item->hasModelType<OffSpecularInstrumentItem>()) {
            m_offSpecularEditor->setItem(dynamic_cast<OffSpecularInstrumentItem*>(m_item));
            m_stack->setCurrentWidget(m_offSpecularWidget);
        } else if (m_item->hasModelType<GISASInstrumentItem>()) {
            m_gisasEditor->setItem(dynamic_cast<GISASInstrumentItem*>(m_item));
            m_stack->setCurrentWidget(m_gisasWidget);
        } else if (m_item->hasModelType<DepthProbeInstrumentItem>()) {
            m_depthProbeEditor->setItem(dynamic_cast<DepthProbeInstrumentItem*>(m_item));
            m_stack->setCurrentWidget(m_depthProbeWidget);
        } else
            ASSERT(false);
    } else {
        m_nameLineEdit->clear();
        m_nameLineEdit->setEnabled(false);
        m_typeLabel->clear();
        m_stack->setCurrentWidget(m_blankEditor);
    }
}

void InstrumentEditor::onInstrumentNameChanged()
{
    if (m_item)
        m_item->setItemName(m_nameLineEdit->text());
}
