//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Realspace/RealSpaceCanvas.cpp
//! @brief     Implements class RealSpaceScene
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Realspace/RealSpaceCanvas.h"
#include "GUI/Application/ApplicationSettings.h"
#include "GUI/Model/Group/FilterPropertyProxy.h"
#include "GUI/Model/Sample/SampleModel.h"
#include "GUI/Model/Session/SessionItemUtils.h"
#include "GUI/Model/State/SessionData.h"
#include "GUI/View/Info/CautionSign.h"
#include "GUI/View/Realspace/RealSpaceBuilder.h"
#include "GUI/ba3d/model/model.h"
#include "GUI/ba3d/view/canvas.h"
#include <QApplication>
#include <QFileDialog>
#include <QMessageBox>
#include <QVBoxLayout>
#include <memory>

RealSpaceCanvas::RealSpaceCanvas(const MaterialModel* materialModel, SampleModel* sampleModel,
                                 QWidget* parent)
    : QWidget(parent)
    , m_sampleModel(sampleModel)
    , m_materialModel(materialModel)
    , m_view(new GUI::RealSpace::Canvas)
    , m_cautionSign(new CautionSign(this))
    , m_currentItem(nullptr)
    , m_firstView(true)
{
    auto* layout = new QVBoxLayout;
    layout->setMargin(0);
    layout->setSpacing(0);
    layout->addWidget(m_view);
    setLayout(layout);
}

void RealSpaceCanvas::setCurrentItem(SessionItem* item)
{
    m_currentItem = item;
    updateScene();
    defaultView(); // Enforces default view and also sets the zoomLevel to default i.e. 0
}

void RealSpaceCanvas::changeLayerSize(double layerSizeChangeScale)
{
    // when no object is selected --> take no action
    if (m_currentItem == nullptr)
        return;

    m_sceneGeometry.layerSize = m_sceneGeometry.layerSize * layerSizeChangeScale;
    updateScene();
}

void RealSpaceCanvas::savePicture()
{
    QPixmap pixmap(this->size());
    render(&pixmap, QPoint(), childrenRegion());
    savePicture(pixmap);
}

void RealSpaceCanvas::onRowsAboutToBeRemoved(const QModelIndex& parent, int first, int)
{
    // clear scene if current selection will be removed
    if (!m_sampleModel)
        return;
    const QModelIndex idx = m_sampleModel->index(first, 0, parent);
    const auto* itemToBeRemoved = m_sampleModel->itemForIndex(idx);
    if (m_currentItem == itemToBeRemoved)
        resetScene();
}

void RealSpaceCanvas::showEvent(QShowEvent*)
{
    setConnected(m_sampleModel, true);
    updateScene();
    if (m_firstView)
        m_view->defaultView();
    m_firstView = false;
}

void RealSpaceCanvas::hideEvent(QHideEvent*)
{
    setConnected(m_sampleModel, false);
}

void RealSpaceCanvas::savePicture(const QPixmap& pixmap)
{
    QString dirname = gSessionData->projectDocument->userExportDir();
    QString defaultExtension = ".png";
    QString selectedFilter("*" + defaultExtension);
    QString defaultName = dirname + "/untitled";
    QString fileName = QFileDialog::getSaveFileName(
        nullptr, "Save 3D real space view", defaultName, selectedFilter, nullptr,
        appSettings->useNativeFileDialog() ? QFileDialog::Options()
                                           : QFileDialog::DontUseNativeDialog);
    QString nameToSave =
        fileName.endsWith(defaultExtension) ? fileName : fileName + defaultExtension;

    if (!nameToSave.isEmpty()) {
        try {
            pixmap.save(nameToSave);
        } catch (const std::exception& ex) {
            QString message = "Attempt to save file with the name '";
            message.append(nameToSave);
            message.append("' has failed with following error message\n\n");
            message.append(QString::fromStdString(ex.what()));
            QMessageBox::warning(nullptr, "Houston, we have a problem.", message);
        }
    }
}

void RealSpaceCanvas::updateScene()
{
    if (!m_view->isValid()) // GL not initialized (widget not shown so far) -> no updating possible
        return;

    QApplication::setOverrideCursor(Qt::WaitCursor);

    m_realSpaceModel = std::make_unique<GUI::RealSpace::Model>();

    RealSpaceBuilder builder3D(m_materialModel);

    try {
        m_cautionSign->clear();
        if (m_currentItem)
            builder3D.populate(m_realSpaceModel.get(), *m_currentItem, m_sceneGeometry,
                               m_view->cam()->getPos());
    } catch (const std::exception& ex) {
        m_cautionSign->setCautionMessage(ex.what());
    } catch (...) {
        // ignore other exceptions thrown
    }
    m_view->setModel(m_realSpaceModel.get());

    QApplication::restoreOverrideCursor();
}

void RealSpaceCanvas::resetScene()
{
    m_realSpaceModel.reset();
    m_view->setModel(nullptr);
    m_currentItem = nullptr;
}

void RealSpaceCanvas::defaultView()
{
    m_view->defaultView();
}

void RealSpaceCanvas::sideView()
{
    m_view->sideView();
}

void RealSpaceCanvas::topView()
{
    m_view->topView();
}

void RealSpaceCanvas::setConnected(SampleModel* model, bool makeConnected)
{
    if (!model)
        return;

    if (makeConnected) {
        connect(model, &SampleModel::rowsInserted, this, &RealSpaceCanvas::updateScene,
                Qt::UniqueConnection);
        connect(model, &SampleModel::rowsRemoved, this, &RealSpaceCanvas::updateScene,
                Qt::UniqueConnection);
        connect(model, &SampleModel::rowsAboutToBeRemoved, this,
                &RealSpaceCanvas::onRowsAboutToBeRemoved, Qt::UniqueConnection);
        connect(model, &SampleModel::itemDataChanged, this, &RealSpaceCanvas::updateScene,
                Qt::UniqueConnection);
        connect(model, &SampleModel::modelReset, this, &RealSpaceCanvas::resetScene,
                Qt::UniqueConnection);
        connect(
            model, &SampleModel::modelAboutToBeReset, this, [&]() { m_currentItem = nullptr; },
            Qt::UniqueConnection);

    } else
        model->disconnect(this);
}
