//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/SampleDesigner/MaterialInplaceForm.cpp
//! @brief     Implements class MaterialInplaceForm
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/SampleDesigner/MaterialInplaceForm.h"
#include "GUI/Model/Material/MaterialItem.h"
#include "GUI/Model/Material/MaterialModel.h"
#include "GUI/Model/Sample/ItemWithMaterial.h"
#include "GUI/Model/Session/ModelPath.h"
#include "GUI/Model/State/SessionData.h"
#include "GUI/Model/Types/DoubleDescriptor.h"
#include "GUI/Model/Types/VectorDescriptor.h"
#include "GUI/View/Edit/DoubleLineEdit.h"
#include "GUI/View/Edit/DoubleSpinBox.h"
#include "GUI/View/Main/MainWindow.h"
#include "GUI/View/MaterialEditor/MaterialEditorDialog.h"
#include "GUI/View/SampleDesigner/LayerEditorUtils.h"
#include "GUI/View/SampleDesigner/SampleEditorController.h"
#include "GUI/View/Tool/LayoutUtils.h"

#include <QGridLayout>
#include <QLabel>
#include <QPushButton>

MaterialInplaceForm::MaterialInplaceForm(QWidget* parent, ItemWithMaterial* item,
                                         SampleEditorController* ec)
    : QWidget(parent), m_item(item), m_ec(ec)
{
    m_layout = new QGridLayout(this);
    m_layout->setContentsMargins(0, 0, 0, 0);
    createWidgets();

    connect(gSessionData->projectDocument->materialModel(), &MaterialModel::materialChanged, this,
            &MaterialInplaceForm::onMaterialChanged);
}

ItemWithMaterial* MaterialInplaceForm::itemWithMaterial() const
{
    return m_item;
}

void MaterialInplaceForm::updateValues()
{
    for (auto* editor : findChildren<DoubleSpinBox*>()) {
        QSignalBlocker b(editor);
        editor->setBaseValue(editor->valueDescriptor());
    }
    for (auto* editor : findChildren<DoubleLineEdit*>()) {
        QSignalBlocker b(editor);
        editor->setBaseValue(editor->valueDescriptor());
    }
}

void MaterialInplaceForm::selectMaterial()
{
    const QString newMaterialIdentifier = MaterialEditorDialog::chooseMaterial(
        baWin, gSessionData->projectDocument, m_item->materialIdentifier());

    if (!newMaterialIdentifier.isEmpty() && newMaterialIdentifier != m_item->materialIdentifier()) {
        GUI::Util::Layout::clearLayout(m_layout, true);
        m_ec->selectMaterial(m_item, newMaterialIdentifier);
        createWidgets();
    } else
        updateValues(); // necessary, since in the material editor the values could have been
                        // changed without selecting a different material
}

void MaterialInplaceForm::createWidgets()
{
    // We can not use the DoubleDescriptors from the MaterialItem itself, because e.g. when editing
    // a material in the material editor, the whole material model will be created from scratch.
    // This means, that all SessionItems will be deleted. If we would use these SessionItems, a
    // crash will happen once an access takes place.
    // What not changes is the material identifier (it is not changed when
    // material is edited in the material editor dialog). Therefore DoubleDescriptors have to be
    // created which use the material identifier for every access.
    // In the following, first the descriptors are copied
    // from the MaterialItem descriptors to get the labels, tooltips etc. Then for set, get &
    // path the functions are redefined to use the material identifier.

    // -- Create UI for delta/beta resp. sldRe/sldIm
    DoubleDescriptors values;
    if (material()->hasRefractiveIndex()) {
        DoubleDescriptor delta = material()->delta();
        delta.set = [=](double value) { material()->delta().set(value); };
        delta.get = [=]() { return material()->delta(); };
        delta.path = [=] { return GUI::Model::Path::getPathFromItem(m_item) + "/delta"; };

        DoubleDescriptor beta = material()->beta();
        beta.set = [=](double value) { material()->beta().set(value); };
        beta.get = [=]() { return material()->beta(); };
        beta.path = [=] { return GUI::Model::Path::getPathFromItem(m_item) + "/beta"; };

        values << delta << beta;
    } else {
        DoubleDescriptor re = material()->sldRe();
        re.set = [=](double value) { material()->sldRe().set(value); };
        re.get = [=]() { return material()->sldRe(); };
        re.path = [=] { return GUI::Model::Path::getPathFromItem(m_item) + "/re"; };

        DoubleDescriptor im = material()->sldIm();
        im.set = [=](double value) { material()->sldIm().set(value); };
        im.get = [=]() { return material()->sldIm(); };
        im.path = [=] { return GUI::Model::Path::getPathFromItem(m_item) + "/im"; };

        values << re << im;
    }

    int col = 0;
    for (const auto& d : values) {
        auto* editor = new DoubleLineEdit(this, d);
        auto* label = new QLabel(d.label, this);
        label->setBuddy(editor);

        QObject::connect(editor, &DoubleLineEdit::baseValueChanged,
                         [=](double newValue) { m_ec->setMaterialValue(m_item, newValue, d); });

        m_layout->addWidget(label, 0, col);
        m_layout->addWidget(editor, 1, col++);
    }

    // -- Create UI for magnetization vector
    VectorDescriptor mag = material()->magnetizationVector();
    mag.x.set = [=](double value) { material()->magnetizationVector().x.set(value); };
    mag.x.get = [=]() { return material()->magnetizationVector().x; };
    mag.x.path = [=] { return GUI::Model::Path::getPathFromItem(m_item) + "/x"; };

    mag.y.set = [=](double value) { material()->magnetizationVector().y.set(value); };
    mag.y.get = [=]() { return material()->magnetizationVector().y; };
    mag.y.path = [=] { return GUI::Model::Path::getPathFromItem(m_item) + "/y"; };

    mag.z.set = [=](double value) { material()->magnetizationVector().z.set(value); };
    mag.z.get = [=]() { return material()->magnetizationVector().z; };
    mag.z.path = [=] { return GUI::Model::Path::getPathFromItem(m_item) + "/z"; };

    const auto setNewValue = [=](double value, DoubleDescriptor d) {
        m_ec->setMaterialValue(m_item, value, d);
    };

    LayerEditorUtils::addVectorToGrid(m_layout, col, mag, setNewValue, true, false);

    // -- Create UI for material selection button
    auto* btn = new QPushButton("...", this);
    btn->setToolTip("Select material");
    m_layout->addWidget(btn, 1, m_layout->columnCount());
    connect(btn, &QPushButton::clicked, this, &MaterialInplaceForm::selectMaterial);

    m_layout->addItem(new QSpacerItem(0, 0, QSizePolicy::Expanding), 0, m_layout->columnCount());
}

MaterialItem* MaterialInplaceForm::material()
{
    return m_item->materialItem();
}

void MaterialInplaceForm::onMaterialChanged(MaterialItem* materialItem)
{
    if (materialItem->identifier() == m_item->materialIdentifier())
        updateValues();
}
