//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Toplevel/ProjectSettingsView.cpp
//! @brief     Implements class ProjectSettingsView
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Toplevel/ProjectSettingsView.h"
#include "GUI/Application/ApplicationSettings.h"
#include "GUI/Model/Instrument/InstrumentModel.h"
#include "GUI/Model/Sample/SampleModel.h"
#include "GUI/Util/Path.h"
#include "GUI/View/Tool/GroupBoxCollapser.h"
#include "ui_ProjectSettingsView.h"
#include <QDir>
#include <QMessageBox>

ProjectSettingsView::ProjectSettingsView(QWidget* parentWidget, ProjectDocument* document)
    : QWidget(parentWidget), m_ui(new Ui::ProjectSettingsView), m_document(document)
{
    m_ui->setupUi(this);
    m_ui->label_6->hide(); // #baTODO show again
    m_ui->textEdit->hide();

    setAttribute(Qt::WA_StyledBackground, true);
    setProperty("stylable", true); // for stylesheet addressing

    GroupBoxCollapser::installIntoGroupBox(m_ui->groupBox);
    GroupBoxCollapser::installIntoGroupBox(m_ui->groupBox_1);
    GroupBoxCollapser::installIntoGroupBox(m_ui->groupBox_2);
    GroupBoxCollapser::installIntoGroupBox(m_ui->groupBox_3);

    connect(m_ui->singleInstrumentModeRadio, &QRadioButton::toggled, this,
            &ProjectSettingsView::onSingleInstrumentRadioToggled);

    connect(m_ui->singleSampleModeRadio, &QRadioButton::toggled, this,
            &ProjectSettingsView::onSingleSampleRadioToggled);

    connect(m_ui->gisasCheck, &QCheckBox::toggled, this,
            &ProjectSettingsView::onFunctionalityToggled);
    connect(m_ui->offSpecCheck, &QCheckBox::toggled, this,
            &ProjectSettingsView::onFunctionalityToggled);
    connect(m_ui->reflectometryCheck, &QCheckBox::toggled, this,
            &ProjectSettingsView::onFunctionalityToggled);
    connect(m_ui->depthCheck, &QCheckBox::toggled, this,
            &ProjectSettingsView::onFunctionalityToggled);

    connect(m_ui->storeAsDefaultsButton, &QPushButton::clicked, this,
            &ProjectSettingsView::storeAsDefaults);

    // #baTODO should be "informationModified" or so?
    connect(m_document, &ProjectDocument::modified, this, &ProjectSettingsView::updateInformation);
}

void ProjectSettingsView::showEvent(QShowEvent*)
{
    writeOptionsToUI();
}

void ProjectSettingsView::writeOptionsToUI()
{
    QSignalBlocker b1(m_ui->singleInstrumentModeRadio);

    m_document->singleInstrumentMode() ? m_ui->singleInstrumentModeRadio->setChecked(true)
                                       : m_ui->multipleInstrumentModeRadio->setChecked(true);

    m_document->singleSampleMode() ? m_ui->singleSampleModeRadio->setChecked(true)
                                   : m_ui->multipleSampleModeRadio->setChecked(true);

    writeFunctionalityToUI();
    updateInformation();
}

void ProjectSettingsView::writeFunctionalityToUI()
{
    const auto f = m_document->functionalities();

    QSignalBlocker b1(m_ui->gisasCheck);
    QSignalBlocker b2(m_ui->offSpecCheck);
    QSignalBlocker b3(m_ui->reflectometryCheck);
    QSignalBlocker b4(m_ui->depthCheck);

    m_ui->gisasCheck->setChecked(f.testFlag(ProjectDocument::Gisas));
    m_ui->offSpecCheck->setChecked(f.testFlag(ProjectDocument::OffSpecular));
    m_ui->reflectometryCheck->setChecked(f.testFlag(ProjectDocument::Specular));
    m_ui->depthCheck->setChecked(f.testFlag(ProjectDocument::DepthProbe));
}

void ProjectSettingsView::updateInformation()
{
    QString location = "[not saved yet]";

    QString name = m_document->projectName();
    if (m_document->hasValidNameAndPath())
        location = GUI::Util::Path::withTildeHomePath(
            QDir::toNativeSeparators(m_document->projectFileName()));

    m_ui->currentProjectName->setText(name);
    m_ui->currentProjectLocation->setText(location);
}

void ProjectSettingsView::storeAsDefaults()
{
    appSettings->setDefaultFunctionalities(toVariant(functionalities()));
    appSettings->setDefaultIsSingleInstrumentMode(m_ui->singleInstrumentModeRadio->isChecked());
    appSettings->setDefaultIsSingleSampleMode(m_ui->singleSampleModeRadio->isChecked());
}

void ProjectSettingsView::onSingleInstrumentRadioToggled(bool newState)
{
    if (newState) {
        if (m_document->instrumentModel()->instrumentItems().size() > 1) {
            QMessageBox::warning(this, "Select single instrument mode",
                                 "This project already contains more than one instrument. Changing "
                                 "this setting is not possible.");

            QSignalBlocker b1(m_ui->singleInstrumentModeRadio);
            m_ui->multipleInstrumentModeRadio->setChecked(true);
            return;
        }

        m_document->setSingleInstrumentMode(true);
    } else
        m_document->setSingleInstrumentMode(false);
}


void ProjectSettingsView::onSingleSampleRadioToggled(bool newState)
{
    if (newState) {
        if (m_document->sampleModel()->multiLayerItems().size() > 1) {
            QMessageBox::warning(this, "Select single sample mode",
                                 "This project already contains more than one sample. Changing "
                                 "this setting is not possible.");

            QSignalBlocker b1(m_ui->singleSampleModeRadio);
            m_ui->multipleSampleModeRadio->setChecked(true);
            return;
        }

        m_document->setSingleSampleMode(true);
    } else
        m_document->setSingleSampleMode(false);
}

void ProjectSettingsView::onFunctionalityToggled()
{
    const auto f = functionalities();

    if (f == ProjectDocument::None) {
        QMessageBox::warning(this, "Select functionality",
                             "You have to select at least one functionality. Changing "
                             "this setting is not possible.");

        writeFunctionalityToUI();
        return;
    }

    m_document->setFunctionalities(f);
}

ProjectDocument::Functionalities ProjectSettingsView::functionalities() const
{
    ProjectDocument::Functionalities f;
    f.setFlag(ProjectDocument::Gisas, m_ui->gisasCheck->isChecked());
    f.setFlag(ProjectDocument::OffSpecular, m_ui->offSpecCheck->isChecked());
    f.setFlag(ProjectDocument::Specular, m_ui->reflectometryCheck->isChecked());
    f.setFlag(ProjectDocument::DepthProbe, m_ui->depthCheck->isChecked());
    return f;
}
